import os
import re
import shutil
import subprocess

import numpy as np
from ase import Atoms
from ase.units import Ang, eV

from pyxtal import pyxtal
from pyxtal.lattice import Lattice

at_types = {
    "C_c": "C1",  #  =  Sp2 C carbonyl group
    "C_cs": "C2",  # =  Sp2 C in c=S
    "C_c1": "C3",  #  Sp C
    "C_c2": "C4",  # Sp2 C
    "C_c3": "C5",  # =  Sp3 C
    "C_ca": "C6",  # Sp2 C in pure aromatic systems
    "C_cp": "C7",  # Head Sp2 C that connect two rings in biphenyl sys.
    "C_cq": "C8",  # Head Sp2 C that connect two rings in biphenyl sys. identical to cp
    "C_cc": "C9",  # Sp2 carbons in non-pure aromatic systems
    "C_cd": "C10",  # Sp2 carbons in non-pure aromatic systems, identical to cc
    "C_ce": "C11",  # Inner Sp2 carbons in conjugated systems
    "C_cf": "C12",  # Inner Sp2 carbons in conjugated systems, identical to ce
    "C_cg": "C13",  # Inner Sp carbons in conjugated systems
    "C_ch": "C14",  # Inner Sp carbons in conjugated systems, identical to cg
    "C_cx": "C15",  # Sp3 carbons in triangle systems
    "C_cy": "C14",  # Sp3 carbons in square systems
    "C_cu": "C16",  # Sp2 carbons in triangle systems
    "C_cv": "C17",  # Sp2 carbons in square systems
    "C_cz": "C18",  # Sp2 carbon in guanidine group
    "H_h1": "H1",  # H bonded to aliphatic carbon with 1 electrwd. group
    "H_h2": "H2",  # H bonded to aliphatic carbon with 2 electrwd. group
    "H_h3": "H3",  # H bonded to aliphatic carbon with 3 electrwd. group
    "H_h4": "H4",  # H bonded to non-sp3 carbon with 1 electrwd. group
    "H_h5": "H5",  # H bonded to non-sp3 carbon with 2 electrwd. group
    "H_ha": "H6",  # H bonded to aromatic carbon
    "H_hc": "H7",  # H bonded to aliphatic carbon without electrwd. group
    "H_hn": "H8",  # H bonded to nitrogen atoms
    "H_ho": "H9",  # Hydroxyl group
    "H_hp": "H10",  # H bonded to phosphate
    "H_hs": "H11",  # Hydrogen bonded to sulphur
    "H_hw": "H12",  # Hydrogen in water
    "H_hx": "H13",  # H bonded to C next to positively charged group
    "F": "F",  # Fluorine
    "Cl": "Cl",  # Chlorine
    "Br": "Br",  # Bromine
    "I": "I",  #  Iodine
    "N_n": "N1",  # Sp2 nitrogen in amide groups
    "N_n1": "N2",  # Sp N
    "N_n2": "N3",  # aliphatic Sp2 N with two connected atoms
    "N_n3": "N4",  # Sp3 N with three connected atoms
    "N_n4": "N5",  # Sp3 N with four connected atoms
    "N_na": "N6",  # Sp2 N with three connected atoms
    "N_nb": "N7",  # Sp2 N in pure aromatic systems
    "N_nc": "N8",  # Sp2 N in non-pure aromatic systems
    "N_nd": "N9",  # Sp2 N in non-pure aromatic systems, identical to nc
    "N_ne": "N10",  # Inner Sp2 N in conjugated systems
    "N_nf": "N11",  # Inner Sp2 N in conjugated systems, identical to ne
    "N_nh": "N12",  # Amine N connected one or more aromatic rings
    "N_no": "N13",  # Nitro N
    "N_ns": "N14",  # amind N, with 1 attached hydrogen atom
    "N_nt": "N15",  # amide N, with 2 attached hydrogen atoms
    "N_nx": "N16",  # like n4, but only has one hydrogen atom
    "N_ny": "N17",  # like n4, but only has two hydrogen atoms
    "N_nz": "N18",  # like n4, but only has three three hydrogen atoms
    "N_n+": "N19",  # NH4+
    "N_nu": "N20",  # like nh, but only has one attached hydrogen atom
    "N_nv": "N21",  # like nh, but only has two attached hydrogen atoms
    "N_n7": "N22",  # like n3, but only has one attached hydrogen atom
    "N_n8": "N23",  # like n3, but only has two attached hydrogen atoms
    "N_n9": "N24",  # NH3
    "O_o": "O1",  # Oxygen with one connected atom
    "O_oh": "O2",  # Oxygen in hydroxyl group
    "O_os": "O3",  # Ether and ester oxygen
    "O_ow": "O4",  # Oxygen in water
    "P_p2": "P1",  # Phosphate with two connected atoms
    "P_p3": "P2",  # Phosphate with three connected atoms, such as PH3
    "P_p4": "P3",  # Phosphate with three connected atoms, such as O=P(CH3)2
    "P_p5": "P4",  # Phosphate with four connected atoms, such as O=P(OH)3
    "P_pb": "P5",  # Sp2 P in pure aromatic systems
    "P_pc": "P6",  # Sp2 P in non-pure aromatic systems
    "P_pd": "P7",  # Sp2 P in non-pure aromatic systems, identical to pc
    "P_pe": "P8",  # Inner Sp2 P in conjugated systems
    "P_pf": "P9",  # Inner Sp2 P in conjugated systems, identical to pe
    "P_px": "P10",  # Special p4 in conjugated systems
    "P_py": "P11",  # Special p5 in conjugated systems
    "S_s": "S1",  # S with one connected atom
    "S_s2": "S2",  # S with two connected atom, involved at least one double bond
    "S_s4": "S3",  # S with three connected atoms
    "S_s6": "S4",  # S with four connected atoms
    "S_sh": "S5",  # Sp3 S connected with hydrogen
    "S_ss": "S6",  # Sp3 S in thio-ester and thio-ether
    "S_sx": "S7",  # Special s4 in conjugated systems
    "S_sy": "S8",  # Special s6 in conjugated systems
}


class GULP:
    """
    A calculator to perform atomic structure optimization in GULP

    Args:

    struc: structure object generated by Pyxtal
    ff: path of forcefield lib
    opt: `conv`, `conp`, `single`
    pstress (float): external pressure
    steps (int): relaxation steps
    symm (bool): whether or not impose the symmetry
    """

    def __init__(
        self,
        struc,
        label="_",
        path="tmp",
        ff="reaxff",
        pstress=None,
        opt="conp",
        steps=1000,
        exe="gulp",
        input="gulp.in",
        output="gulp.log",
        dump=None,
        symmetry=False,
        labels=None,
        timeout=3600,
    ):
        if isinstance(struc, pyxtal):
            self.pyxtal = struc
            self.species = struc.species
            struc = struc.to_ase(resort=False)
        else:
            self.pyxtal = None

        if isinstance(struc, Atoms):
            self.lattice = Lattice.from_matrix(struc.cell)
            self.frac_coords = struc.get_scaled_positions()
            self.sites = struc.get_chemical_symbols()
            self.species = None
        else:
            raise NotImplementedError("only support ASE atoms object")

        self.symmetry = symmetry  # ; print(self.pyxtal.lattice.ltype)
        self.structure = struc
        self.pstress = pstress
        self.label = label
        self.labels = labels
        self.ff = ff
        self.opt = opt
        self.exe = exe
        self.steps = steps
        self.folder = path
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        self.input  = self.label + input
        self.output = self.label + output
        self.dump = dump
        self.iter = 0
        self.energy = None
        self.energy_per_atom = None
        self.stress = None
        self.forces = None
        self.positions = None
        self.optimized = False
        self.cputime = 0
        self.error = False
        self.timeout = timeout

    def set_catlow(self):
        """
        set the atomic label for catlow potentials
        O_O2- general O2- species
        O_OH oxygen in hydroxyl group
        H_OH hydrogen in hydroxyl group
        """

    def run(self, clean=True):
        """
        Always go to the directory to run one gulp at once
        """
        cwd = os.getcwd()
        os.chdir(self.folder)
        self.write()
        self.execute()
        self.read()
        if clean:
            self.clean()
        os.chdir(cwd)

    def execute(self):
        cmd = self.exe + "<" + self.input + ">" + self.output
        # os.system(cmd)
        with open(os.devnull, 'w') as devnull:
            try:
                # Run the external command with a timeout
                result = subprocess.run(
                    cmd, shell=True, timeout=self.timeout, check=True, stderr=devnull)
                return result.returncode  # Or handle the result as needed
            except subprocess.CalledProcessError as e:
                print(f"Command '{cmd}' failed with return code {e.returncode}.")
                return None
            except subprocess.TimeoutExpired:
                print(f"External command {cmd} timed out.")
                return None

    def clean(self):
        if self.error:
            os.system("mv " + self.input + " " + self.input + "_error")
            os.system("mv " + self.output + " " + self.output + "_error")
        else:
            os.remove(self.input)
            os.remove(self.output)
        if self.dump is not None:
            os.remove(self.dump)

    def to_ase(self):
        return Atoms(self.sites, scaled_positions=self.frac_coords, cell=self.lattice.matrix)

    def to_pymatgen(self):
        from pymatgen.core.structure import Structure

        return Structure(self.lattice.matrix, self.sites, self.frac_coords)

    def to_pyxtal(self):
        ase_atoms = self.to_ase()
        for tol in [1e-2, 1e-3, 1e-4, 1e-5]:
            try:
                struc = pyxtal()
                struc.from_seed(ase_atoms, tol=tol)
                break
            except:
                pass
                # print('Something is wrong', tol)
                # struc.from_seed('bug.vasp', tol*10)
        return struc

    def write(self):
        a, b, c, alpha, beta, gamma = self.lattice.get_para(degree=True)

        with open(self.input, "w") as f:
            if self.opt == "conv":
                f.write(f"opti stress {self.opt:s} conjugate ")
            elif self.opt == "single":
                f.write("grad conp stress ")
            else:
                f.write(f"opti stress {self.opt:s} conjugate ")

            if not self.symmetry:
                f.write("nosymmetry\n")

            f.write("\ncell\n")
            f.write(f"{a:12.6f}{b:12.6f}{c:12.6f}{alpha:12.6f}{beta:12.6f}{gamma:12.6f}\n")
            f.write("\nfractional\n")

            if self.symmetry and self.pyxtal is not None:
                # Use pyxtal here
                for site in self.pyxtal.atom_sites:
                    symbol, coord = site.specie, site.position
                    f.write("{:4s} {:12.6f} {:12.6f} {:12.6f} core \n".format(symbol, *coord))
                    if self.ff == "catlow" and symbol == "O":
                        f.write("{:4s} {:12.6f} {:12.6f} {:12.6f} shell \n".format(symbol, *coord))

                # Tested for all space groups
                f.write(f"\nspace\n{self.pyxtal.group.number:d}\n")
                f.write("\norigin\n0 0 0\n")
            else:
                # All coordinates
                for coord, site in zip(self.frac_coords, self.sites):
                    f.write("{:4s} {:12.6f} {:12.6f} {:12.6f} core \n".format(site, *coord))
            species = self.structure.species if self.species is not None else list(set(self.sites))

            f.write("\nSpecies\n")
            if self.labels is not None:
                for specie in species:
                    if specie in self.labels:
                        sp = self.labels[specie]
                        f.write(f"{specie:4s} core {sp:s}\n")
                    else:
                        f.write(f"{specie:4s} core {specie:4s}\n")
            else:
                for specie in species:
                    if self.ff == "catlow" and specie == "O":
                        f.write("O    core O_O2- core\n")
                        f.write("O    shell O_O2- shell\n")
                    else:
                        f.write(f"{specie:4s} core {specie:4s}\n")

            f.write(f"\nlibrary {self.ff:s}\n")
            f.write("ewald 10.0\n")
            # f.write('switch rfo gnorm 1.0\n')
            # f.write('switch rfo cycle 0.03\n')
            if self.opt != "single":
                f.write(f"maxcycle {self.steps:d}\n")
            if self.pstress is not None:
                f.write(f"pressure {self.pstress:6.3f}\n")
            if self.dump is not None:
                f.write(f"output cif {self.dump:s}\n")

    def read(self):
        # for symmetry case
        lattice_para = None
        lattice_vector = None
        ltype = self.pyxtal.lattice.ltype if self.pyxtal is not None else "triclinic"

        with open(self.output) as f:
            lines = f.readlines()
        try:
            for i, line in enumerate(lines):
                if self.symmetry and self.pyxtal.group.symbol[0] != "P":
                    m = re.match(r"\s*Non-primitive unit cell\s*=\s*(\S+)\s*eV", line)
                elif self.pstress is None or self.pstress == 0:
                    m = re.match(r"\s*Total lattice energy\s*=\s*(\S+)\s*eV", line)
                else:
                    m = re.match(r"\s*Total lattice enthalpy\s*=\s*(\S+)\s*eV", line)
                # print(line.find('Final asymmetric unit coord'), line)
                if m:
                    self.energy = float(m.group(1))
                    self.energy_per_atom = self.energy / len(self.frac_coords)

                elif line.find("Job Finished") != -1:
                    self.optimized = True

                elif line.find("Total CPU time") != -1:
                    self.cputime = float(line.split()[-1])

                elif line.find("Final stress tensor components") != -1:
                    stress = np.zeros([6])
                    for j in range(3):
                        var = lines[i + j + 3].split()[1]
                        stress[j] = float(var)
                        var = lines[i + j + 3].split()[3]
                        stress[j + 3] = float(var)
                    self.stress = stress

                # Forces, QZ copied from https://gitlab.com/ase/ase/-/blob/master/ase/calculators/gulp.py
                elif line.find("Final internal derivatives") != -1:
                    s = i + 5
                    forces = []
                    while True:
                        s = s + 1
                        if lines[s].find("------------") != -1:
                            break
                        g = lines[s].split()[3:6]

                        for _t in range(3 - len(g)):
                            g.append(" ")
                        for j in range(2):
                            min_index = [i + 1 for i, e in enumerate(g[j][1:]) if e == "-"]
                            if j == 0 and len(min_index) != 0:
                                if len(min_index) == 1:
                                    g[2] = g[1]
                                    g[1] = g[0][min_index[0] :]
                                    g[0] = g[0][: min_index[0]]
                                else:
                                    g[2] = g[0][min_index[1] :]
                                    g[1] = g[0][min_index[0] : min_index[1]]
                                    g[0] = g[0][: min_index[0]]
                                    break
                            if j == 1 and len(min_index) != 0:
                                g[2] = g[1][min_index[0] :]
                                g[1] = g[1][: min_index[0]]

                        G = [-float(x) * eV / Ang for x in g]
                        forces.append(G)
                    forces = np.array(forces)
                    self.forces = forces

                elif line.find(" Cycle: ") != -1:
                    self.iter = int(line.split()[1])

                # asymmetric unit
                elif line.find("Final asymmetric unit coordinates") != -1:
                    s = i + 6
                    positions = []
                    for _i in range(len(self.pyxtal.atom_sites)):
                        xyz = lines[s + _i].split()[3:6]
                        XYZ = [float(x) for x in xyz]
                        # print(XYZ)
                        self.pyxtal.atom_sites[_i].update(XYZ)

                elif line.find("Final fractional coordinates of atoms") != -1:
                    s = i + 5
                    positions = []
                    species = []
                    while True:
                        s = s + 1
                        if lines[s].find("------------") != -1:
                            break
                        xyz = lines[s].split()[3:6]
                        XYZ = [float(x) for x in xyz]
                        positions.append(XYZ)
                        species.append(lines[s].split()[1])
                    # if len(species) != len(self.sites):
                    #    print("Warning", len(species), len(self.sites))
                    self.frac_coords = np.array(positions)
                elif line.find("Final Cartesian lattice vectors") != -1:
                    lattice_vectors = np.zeros((3, 3))
                    s = i + 2
                    for j in range(s, s + 3):
                        temp = lines[j].split()
                        for k in range(3):
                            lattice_vectors[j - s][k] = float(temp[k])
                    lattice_vector = Lattice.from_matrix(lattice_vectors, ltype=ltype)

                elif line.find("Non-primitive lattice parameters") != -1:
                    s = i + 2
                    temp = lines[s].split()
                    a, b, c = float(temp[2]), float(temp[5]), float(temp[8])
                    temp = lines[s + 1].split()
                    alpha, beta, gamma = float(temp[1]), float(temp[3]), float(temp[5])
                    lattice_para = Lattice.from_para(a, b, c, alpha, beta, gamma, ltype)
        except:
            self.error = True
            self.energy = None
        if lattice_para is not None:
            self.lattice = lattice_para
        elif lattice_vector is not None:
            self.lattice = lattice_vector
        else:
            self.error = True
            self.energy = None

        if self.pyxtal is not None:
            self.pyxtal.lattice = self.lattice

        if self.energy is None or np.isnan(self.energy):
            self.error = True
            self.energy = None
            print("GULP calculation is wrong, reading------")


class GULP_OC:
    """
    A calculator to perform oragnic crystal structure optimization in GULP

    Args:
        struc: structure object generated by Pyxtal
        ff: path of forcefield lib, e.g., `gaff`
        bond_type: specify the bond type or not (GULP can detect it automatically)
        opt: e.g., `conv`, `conp`, `single`
        steps: number of steps, int, e.g., `1000`
        exe: int, the way to call gulp executable
        atom_info: atomic labels/charges
        input: gulp input file name
        output: gulp output file name
        dump: dumped gulp file name
    """

    def __init__(
        self,
        struc,
        label="_",
        ff="gaff2",
        bond_type=False,
        opt="conp",
        steps=1000,
        stepmx=0.001,
        exe="gulp",
        atom_info=None,
        input="gulp.in",
        output="gulp.log",
        dump=None,
        folder=".",
    ):
        self.folder = folder
        self.structure = struc
        self.label = label
        self.ff = ff
        self.bond_type = bond_type
        self.opt = opt
        self.exe = exe
        if "OCSP_GULPEXE" in os.environ:
            self.exe = os.environ["OCSP_GULPEXE"]
        self.steps = steps
        self.stepmx = stepmx
        self.opt = opt
        self.input = self.label + input
        self.output = self.label + output
        self.dump = dump
        self.iter = 0
        self.stress = None
        self.positions = None
        self.forces = None
        self.optimized = False
        self.cell = None
        self.group = self.structure.group
        self.optlat = False
        self.cputime = 0
        self.atom_info = atom_info

    def run(self, clean=True, pause=False):
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        cwd = os.getcwd()
        os.chdir(self.folder)

        self.write()
        if not pause:
            self.execute()
            self.read()
            if clean:
                self.clean()

        os.chdir(cwd)

    def execute(self):
        if self.exe == "gulp.exe":
            key = self.input.split(".")[0]
            shutil.copy(self.input, key + ".gin")
            cmd = self.exe + " " + key
            os.system(cmd)
            shutil.copy(key + ".gout", self.output)
        else:
            cmd = self.exe + "<" + self.input + ">" + self.output
            os.system(cmd)

    def clean(self):
        os.remove(self.input)
        os.remove(self.output)

    def write(self):
        lat = self.structure.lattice
        a, b, c = lat.a, lat.b, lat.c
        alpha, beta, gamma = (
            np.degrees(lat.alpha),
            np.degrees(lat.beta),
            np.degrees(lat.gamma),
        )
        ltype = lat.ltype

        with open(self.input, "w") as f:
            # f.write('opti stress {:s} conj molecule nomod qok\n'.format(self.opt))
            if self.opt == "conv":
                # f.write('opti {:s} steepest molecule nomod qok\n'.format(self.opt))
                f.write(f"opti {self.opt:s} conj molecule nomod qok\n")
            else:
                f.write(f"opti stress {self.opt:s} conj molecule nomod qok\n")
            f.write("\ncell\n")
            f.write(f"{a:12.6f}{b:12.6f}{c:12.6f}{alpha:12.6f}{beta:12.6f}{gamma:12.6f}\n")
            f.write("\nfractional\n")

            symbols = []

            for i, site in enumerate(self.structure.mol_sites):
                coords, species = site._get_coords_and_species(first=True)
                if self.atom_info is None:
                    labels = site.molecule.props["gmx_label"]
                    charges = site.molecule.props["charge"]
                else:
                    labels = self.atom_info["label"][site.type]
                    charges = self.atom_info["charge"][site.type]
                for j, coord in enumerate(coords):
                    # print(len(site.molecule.mol.sites), len(labels), len(coords), coord)
                    symbol = site.molecule.mol.sites[j].species_string
                    if symbol not in ["F", "Cl", "Br"]:
                        symbol += "_" + labels[j]
                    symbols.append(symbol)
                    try:
                        f.write(
                            f"{at_types[symbol]:4s} {coord[0]:12.6f} {coord[1]:12.6f} {coord[2]:12.6f} core {charges[j]:12.6f}\n"
                        )
                    except KeyError:
                        msg = f"symbol {symbol:s} is not supported in GULP-gaff2.lib"
                        raise KeyError(msg)

            f.write("\nSpecies\n")
            symbols = list(set(symbols))
            for symbol in symbols:
                f.write(f"{at_types[symbol]:4s} core {symbol:4s}\n")

            # symmetry operations
            f.write(f"\nsymmetry_cell {ltype:s}\n")
            site0 = self.structure.mol_sites[0]
            for op in site0.wp.ops[1:]:
                f.write("symmetry_operator\n")
                rot = op.rotation_matrix.T
                trans = op.translation_vector
                for i in range(3):
                    f.write("{:6.3f} {:6.3f} {:6.3f} {:6.3f}\n".format(*rot[i, :], trans[i]))

            # bond type
            if self.bond_type:
                for bond in site0.mol.molTopol.bonds:
                    for i in range(len(site.wp)):
                        count = i * len(coords)
                        f.write(f"connect {bond.atoms[0].id + count:4d} {bond.atoms[1].id + count:4d}\n")
            f.write(f"\nlibrary {self.ff:s}.lib\n")
            # f.write('\nlibrary {:s}\n'.format(self.ff))
            if "OCSP_GULP_LIB" in os.environ:
                shutil.copy(os.environ["OCSP_GULP_LIB"] + "/" + self.ff + ".lib", ".")
            f.write("ewald 10.0\n")
            # f.write('switch rfo gnorm 1.0\n')
            # f.write('switch lbfgs cycle 300\n')
            f.write(f"maxcycle {self.steps:d}\n")
            f.write("stepmx " + str(self.stepmx) + "\n")
            f.write("ftol 0.0001\n")  # energy tol default: 1e-5
            f.write("gtol 0.002\n")  # force tol default: 1e-4
            f.write("gmax 0.01\n")  # force tol defult: 1e-3
            if self.dump is not None:
                f.write(f"output cif {self.dump:s}\n")
            # https://gulp.curtin.edu.au/gulp/help/new_help_40_txt.html#gmax

    def read(self):
        with open(self.output) as f:
            lines = f.readlines()

        for i, line in enumerate(lines):
            if i == 7:
                self.version = line

            if line.find("Total lattice energy") >= 0:
                m = re.match(r"\s*Total lattice energy\s*=\s*(\S+)\s*eV", line)
                if m:
                    self.structure.energy = float(m.group(1))
            else:
                if line.find("Job Finished") != -1:
                    self.optimized = True

                elif line.find("Total CPU time") != -1:
                    self.cputime = float(line.split()[-1])

                elif line.find("Final stress tensor components") != -1:
                    stress = np.zeros([6])
                    for j in range(3):
                        var = lines[i + j + 3].split()[1]
                        stress[j] = float(var)
                        var = lines[i + j + 3].split()[3]
                        stress[j + 3] = float(var)
                    self.stress = stress

                elif line.find(" Cycle: ") != -1:
                    self.iter = int(line.split()[1])

                elif (
                    line.find("Final asymmetric unit coord") != -1
                    or line.find("Final fractional coordinates of atoms") != -1
                ):
                    s = i + 5
                    positions = []
                    species = []
                    while True:
                        s = s + 1
                        if lines[s].find("------------") != -1:
                            break
                        # if lines[s].find(" s ") != -1:
                        #    continue
                        xyz = lines[s].split()[3:6]
                        XYZ = [float(x) for x in xyz]
                        positions.append(XYZ)
                        species.append(lines[s].split()[1])
                    self.positions = np.array(positions)
                    self.species = species

                elif line.find("Final Cartesian lattice vectors") != -1:
                    lattice_vectors = np.zeros((3, 3))
                    s = i + 2
                    for j in range(s, s + 3):
                        temp = lines[j].split()
                        for k in range(3):
                            lattice_vectors[j - s][k] = float(temp[k])
                    self.cell = lattice_vectors
        if self.cell is None:
            self.cell = self.structure.lattice.matrix

        self.lattice = Lattice.from_matrix(self.cell)  # update the lattice
        self.lattice.ltype = self.structure.group.lattice_type
        self.structure.lattice = self.lattice

        if self.optimized:
            count = 0

            try:
                for site in self.structure.mol_sites:
                    coords = self.positions[count : count + len(site.molecule.mol)]
                    site.update(coords, self.lattice)
                    count += len(site.molecule.mol)
                self.structure.optimize_lattice()
            except:
                # print()
                print("Structure is wrong after optimization")


def single_optimize(
    struc,
    ff,
    steps=1000,
    pstress=None,
    opt="conp",
    exe="gulp",
    path="tmp",
    label="_",
    clean=True,
    symmetry=False,
    labels=None,
):
    calc = GULP(
        struc,
        steps=steps,
        label=label,
        path=path,
        pstress=pstress,
        ff=ff,
        opt=opt,
        symmetry=symmetry,
        labels=labels,
    )

    calc.run(clean=clean)

    if calc.error:
        print("GULP error in single optimize")
        return None, None, 0, True
    else:
        struc = calc.to_pyxtal() if calc.pyxtal is None else calc.pyxtal
        # if sum(struc.numIons) == 42: print("SSSSS"); import sys; sys.exit()
        return struc, calc.energy_per_atom, calc.cputime, calc.error


def optimize(
    struc,
    ff,
    optimizations=None,
    exe="gulp",
    pstress=None,
    path="tmp",
    label="_",
    clean=True,
    adjust=False,
):
    """
    Multiple calls

    """
    if optimizations is None:
        optimizations = ["conp", "conp"]
    time_total = 0
    for opt in optimizations:
        struc, energy, time, error = single_optimize(
            struc,
            ff,
            pstress=pstress,
            opt=opt,
            exe=exe,
            path=path,
            label=label,
            clean=clean,
        )

        time_total += time
        if error:
            return None, None, 0, True
        elif adjust and abs(energy) < 1e-8:
            matrix = struc.lattice.matrix
            struc.lattice.set_matrix(matrix * 0.8)

    return struc, energy, time_total, False


if __name__ == "__main__":
    while True:
        struc = pyxtal()
        struc.from_random(3, 19, ["C"], [4])
        if struc.valid:
            break
    print(struc)
    pmg1 = struc.to_pymatgen()
    calc = GULP(struc, opt="single", ff="tersoff.lib")
    calc.run(clean=False)  # ; import sys; sys.exit()
    print(calc.energy)
    print(calc.stress)
    print(calc.forces)
    pmg2 = calc.to_pymatgen()
    # xtal = calc.pyxtal #calc.to_pyxtal()
    # print(calc.iter)
    # print(xtal)

    import pymatgen.analysis.structure_matcher as sm

    print(sm.StructureMatcher().fit(pmg1, pmg2))

    struc, eng, time, _ = optimize(struc, ff="tersoff.lib")
    print(struc)
    print(eng)
    print(time)



"""
Module for crystal packing descriptor from energy decomposition
"""

import numpy as np
from scipy.optimize import minimize
from scipy.spatial.transform import Rotation
from scipy.special import sph_harm
from scipy.stats import qmc


def _qlm(dists, l=4):
    """
    Calculates the vector associated with an atomic site and
    one of its neighbors

    Args:
        distss: a list of distance vectors
        l:  free integer quantum number
    Returns:
        q: numpy array(complex128), the complex vector qlm normalized
            by the number of nearest neighbors
    """
    # initiate variable as a complex number
    q = np.zeros(2 * l + 1, dtype=np.complex128)
    neighbors_count = len(dists)

    for i, m in enumerate(range(-l, l + 1)):
        for _j, r_vec in enumerate(dists):
            # find the position vector of the site/neighbor pair
            r_mag = np.linalg.norm(r_vec)
            theta = np.arccos(r_vec[2] / r_mag)
            if abs((r_vec[2] / r_mag) - 1.0) < 10.0 ** (-8.0):
                theta = 0.0
            elif abs((r_vec[2] / r_mag) + 1.0) < 10.0 ** (-8.0):
                theta = np.pi

            # phi
            if r_vec[0] < 0.0:
                phi = np.pi + np.arctan(r_vec[1] / r_vec[0])
            elif r_vec[0] > 0.0 and r_vec[1] < 0.0:
                phi = 2 * np.pi + np.arctan(r_vec[1] / r_vec[0])
            elif r_vec[0] > 0.0 and r_vec[1] >= 0.0:
                phi = np.arctan(r_vec[1] / r_vec[0])
            elif r_vec[0] == 0.0 and r_vec[1] > 0.0:
                phi = 0.5 * np.pi
            elif r_vec[0] == 0.0 and r_vec[1] < 0.0:
                phi = 1.5 * np.pi
            else:
                phi = 0.0

            q[i] += sph_harm(m, l, phi, theta)
    # normalize by number of neighbors
    return q / neighbors_count


def correlation(coef1, coef2, angle=None, s=0):
    """
    Compute the correlation between to sph coefs

    Args:
        coef1: sph coefficients 1
        coef2: sph coefficients 2
        angle: [alpha, beta, gamma]
        s: starting index of coefs

    Return:
        distance scaled in [0, 1]
    """
    if angle is not None:
        coef2 = coef2.rotate(angle[0], angle[1], angle[2], degrees=True)
    power = np.sqrt(coef1.spectrum()[s:].sum() * coef2.spectrum()[s:].sum())
    cross = coef1.cross_spectrum(coef2)[s:]
    return cross.sum() / power


def correlation_opt(coef1, coef2, angle, s=0):
    """
    Compute the correlation between two sph coefs

    Args:
        coef1: sph coefficients 1
        coef2: sph coefficients 2
        angle: [alpha, beta, gamma]
        s: starting index of coefs

    Return:
        distance scaled in [0, 1]
    """

    def fun(x0, coef1, coef2, s):
        coef = coef2.rotate(x0[0], x0[1], x0[2], degrees=True)
        return -correlation(coef1, coef, s=s)

    res = minimize(
        fun,
        angle,
        args=(coef1, coef2, s),
        method="Nelder-Mead",
        options={"maxiter": 20},
    )
    return -res.fun, res.x


def correlation_go(coef1, coef2, M=6, s=0, d_cut=0.92):
    """
    global optimization of two coefs based on quasi random sampling

    Args:
        coef1: sph coefficients 1
        coef2: sph coefficients 2
        M: 2^M sampling points
        s: starting index of coefs

    Return:
        distance scaled in [0, 1]

    """
    sampler = qmc.Sobol(d=3, scramble=False)
    sample = sampler.random_base2(m=M)
    sample = qmc.scale(sample, [-180, -90, -180], [180, 90, 180])

    ds, angles = [], []
    for angle in sample:
        d, angle = correlation_opt(coef1, coef2, angle, s=0)
        ds.append(d)
        angles.append(angle)
        if d > d_cut * 1.1:
            break
    ds = np.array(ds)
    id = np.argmax(ds)
    return ds[id], angles[id]


def fibonacci_sphere(N=1000):
    """
    Sampling the sphere grids

    Args:
        N: number of pts to generate

    Returns:
        3D points array in Cartesian coordinates
    """
    points = []
    phi = np.pi * (3.0 - np.sqrt(5.0))  # golden angle in radians
    for i in range(N):
        y = 1 - (i / float(N - 1)) * 2  # y goes from 1 to -1
        radius = np.sqrt(1 - y * y)  # radius at y
        theta = phi * i  # golden angle increment
        x = np.cos(theta) * radius
        z = np.sin(theta) * radius
        points.append((x, y, z))

    return np.array(points)


def cart2sph(x, y, z):
    """
    convert the x, y, z to spherical coordinates (phi, theta, r)
    phi: [-pi, pi]
    theta: [-pi/2, pi/2]
    """
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    theta = np.arctan2(z, hxy)
    phi = np.arctan2(y, x)
    return phi, theta, r


def sph2cart(phi, theta, r):
    """
    convert spherical coordinates (phi, theta, r) to Cartesian (x, y, z)
    """
    rcos_theta = r * np.cos(theta)
    x = rcos_theta * np.cos(phi)
    y = rcos_theta * np.sin(phi)
    z = r * np.sin(theta)
    return x, y, z


def xyz2sph(xyzs, radian=True):
    """
    convert the vectors (x, y, z) to the sphere representation (theta, phi)

    Args:
        xyzs: 3D xyz coordinates
        radian: return in radian (otherwise degree)
    """
    pts = np.zeros([len(xyzs), 2])
    for i, r_vec in enumerate(xyzs):
        r_mag = np.linalg.norm(r_vec)
        theta0 = np.arccos(r_vec[2] / r_mag)
        if abs((r_vec[2] / r_mag) - 1.0) < 10.0 ** (-8.0):
            theta0 = 0.0
        elif abs((r_vec[2] / r_mag) + 1.0) < 10.0 ** (-8.0):
            theta0 = np.pi

        if r_vec[0] < 0.0:
            phi0 = np.pi + np.arctan(r_vec[1] / r_vec[0])
        elif r_vec[0] > 0.0 and r_vec[1] < 0.0:
            phi0 = 2 * np.pi + np.arctan(r_vec[1] / r_vec[0])
        elif r_vec[0] > 0.0 and r_vec[1] >= 0.0:
            phi0 = np.arctan(r_vec[1] / r_vec[0])
        elif r_vec[0] == 0.0 and r_vec[1] > 0.0:
            phi0 = 0.5 * np.pi
        elif r_vec[0] == 0.0 and r_vec[1] < 0.0:
            phi0 = 1.5 * np.pi
        else:
            phi0 = 0.0
        pts[i, :] = [theta0, phi0]
    if not radian:
        pts = np.degree(pts)

    return pts


def expand_sph(pts, l_max, norm=4, csphase=-1):
    from pyshtools.expand import SHExpandLSQ

    """
    Transform the grid points to spherical harmonics

    Args:
        pts: 3D array
            (thetas, phis, vals) with a length of N

        lmax: Integer
            The maximum degree of spherical harmonic.

        coeff_norm: integer
            The normalization of SHExpandLSQ().
            1 (default) = Geodesy 4-pi normalized harmonics;
            2 = Schmidt semi-normalized harmonics;
            3 = unnormalized harmonics;
            4 = orthonormal harmonics.

        csphase: Integer
            whether (-1) or not (1) apply the Condon-Shortley phase factor.

    Return:
        cilm: float, dimension (2, lmax+1, lmax+1)
            Coefficients of the spherican harmonic

        chi2: float
            The residual sum of squares misfit for an overdetermined inversion.
    """
    thetas, phis, vals = pts[:, 0], pts[:, 1], pts[:, 2]
    phis = np.degrees(phis)
    # shift from (0, 180) to (-90, 90)
    thetas = np.degrees(thetas)
    thetas -= 90

    # if thetas is within [0, 180]
    # print('check thetas', thetas.min(), thetas.max())
    # if abs(thetas.min()) < 1e-3: thetas -= 90
    cilm, chi2 = SHExpandLSQ(vals, thetas, phis, l_max, norm=norm, csphase=csphase)

    return cilm, chi2


def get_alignment(pts, degrees=True):
    """
    Here we define the equator is the plane with three most important neighbors.
    Get the required rotation angles to get that representation.

    Args:
        pts; important points (3D array)

    Returns:
        angles: [alpha, beta, gamma]
    """
    # get the three most importants ids
    tps = pts[:3]

    xyz0 = np.array(
        [
            np.sin(tps[0, 0]) * np.cos(tps[0, 1]),
            np.sin(tps[0, 1]) * np.sin(tps[0, 0]),
            np.cos(tps[0, 0]),
        ]
    )
    xyz1 = np.array(
        [
            np.sin(tps[1, 0]) * np.cos(tps[1, 1]),
            np.sin(tps[1, 1]) * np.sin(tps[1, 0]),
            np.cos(tps[1, 0]),
        ]
    )
    xyz2 = np.array(
        [
            np.sin(tps[2, 0]) * np.cos(tps[2, 1]),
            np.sin(tps[2, 1]) * np.sin(tps[2, 0]),
            np.cos(tps[2, 0]),
        ]
    )
    line1, line2 = xyz1 - xyz0, xyz2 - xyz0
    ax1 = np.cross(line1, line2)
    ax1 /= np.linalg.norm(ax1)
    ax0 = np.array([0.0, 0.0, 1.0])
    #
    vec = np.cross(ax0, ax1)
    angle = np.arccos(np.dot(ax0, ax1))
    r = Rotation.from_rotvec(angle * vec)
    angles = r.as_euler("zyz")
    if degrees:
        angles = np.degrees(angles)
    return angles


class spherical_image:
    import pyshtools as pysh

    """
    A class to handle the crystal packing descriptor from spherical image

    Args:
        xtal: pyxtal structure
        model: 'molecule' or 'contact'
        max_d: maximum intermolecular distances
        lmax: maximum bandwidth for spherical harmonic expansion
        sigma: Gaussian width to project into the unit sphere
        N: number of grid points on the unit sphere
    """

    def __init__(self, xtal, model="molecule", max_d=10, factor=2.2, lmax=13, sigma=0.1, N=10000):
        for i in range(len(xtal.mol_sites)):
            try:
                numbers = xtal.mol_sites[i].molecule.mol.atomic_numbers
                if numbers.count(7) > 0 or numbers.count(8) > 0:
                    xtal.mol_sites[i].molecule.set_labels()
            except:
                print("Warning! Needs the smiles information!")
        self.xtal = xtal
        self.max_d = max_d
        self.factor = factor
        self.lmax = lmax
        self.sigma = sigma
        self.N = N

        xyzs = fibonacci_sphere(N)
        grids = np.zeros([N, 3])
        grids[:, :2] = xyz2sph(xyzs)

        if model == "molecule":
            self.pts = self.get_molecules()
        else:
            self.pts = self.get_contacts()
        self.coefs = []
        for pt in self.pts:
            grids[:, 2] = self.calculate_density(pt, xyzs)
            cilm, chi2 = expand_sph(grids, lmax)
            self.coefs.append(pysh.SHCoeffs.from_array(cilm))
        self.ds = np.ones(len(self.coefs))

    def calculate_density(self, pt, xyzs):
        """
        calculate the projected density on the unit sphere
        """
        vals = np.zeros(len(xyzs))
        for _pt in pt:
            t0, p0, h = _pt
            x0, y0, z0 = np.sin(t0) * np.cos(p0), np.sin(t0) * np.sin(p0), np.cos(t0)
            dst = np.linalg.norm(xyzs - np.array([x0, y0, z0]), axis=1)
            vals += h * np.exp(-(dst**2 / (2.0 * self.sigma**2)))
        return vals

    def get_molecules(self):
        """
        compute the spherical images from neighboring molecules

        Returns:
            pts: [N, 3] array, (theta, phi, eng)
        """
        pts = []
        for i, site in enumerate(self.xtal.mol_sites):
            _, neighs, comps, _, engs = self.xtal.get_neighboring_molecules(
                i, factor=self.factor, max_d=self.max_d, ignore_E=False
            )
            xyz, _ = site._get_coords_and_species(absolute=True, first=True)
            center = site.molecule.get_center(xyz)
            coords = np.zeros([len(neighs), 3])
            for _i, xyz in enumerate(neighs):
                # coords[_i, :] = site.molecule.get_center(xyz) - center
                coords[_i, :] = self.xtal.molecules[comps[_i]].get_center(xyz) - center
            pt = np.zeros([len(coords), 3])
            pt[:, :2] = xyz2sph(coords)
            pt[:, 2] = engs / np.sum(engs)
            pts.append(pt)
        return pts

    def get_contacts(self):
        """
        Compute the spherical images from the neighboring distances

        Returns:
            pts: [N, 3] array, (theta, phi, eng)
        """
        pts = []
        for i, _site in enumerate(self.xtal.mol_sites):
            engs, pairs, dists = self.xtal.get_neighboring_dists(i, factor=self.factor, max_d=self.max_d)
            pt = np.zeros([len(pairs), 3])
            pt[:, :2] = xyz2sph(pairs)
            pt[:, 2] = engs / np.sum(engs)
            pts.append(pt)
        return pts

    def plot_sph_images(self, lmax=None, figname=None, molecule=False):
        """
        Plot the spherical images in both 3d and 2d

        Args:
            lmax: maximum truncation
            figname: name of figure file
            molecule: draw 2D molecule diagram or not
        """
        import matplotlib.gridspec as gridspec
        import matplotlib.pyplot as plt

        if molecule:
            nrows = len(self.coefs) + 1
            shift = 1
        else:
            nrows = len(self.coefs)
            shift = 0

        fig = plt.figure(figsize=(9, 4 * nrows))
        gs = gridspec.GridSpec(nrows=nrows, ncols=2, wspace=0.15, width_ratios=[0.7, 1])
        if molecule:
            from rdkit import Chem
            from rdkit.Chem import Draw

            smi = ""
            for i, m in enumerate(self.xtal.molecules):
                smi += m.smile
                if i + 1 < len(self.xtal.molecules):
                    smi += "."

            m = Chem.MolFromSmiles(smi)
            im = Draw.MolToImage(m)
            ax0 = fig.add_subplot(gs[0, :])
            plt.imshow(im)
            ax0.axis("off")

        if lmax is None:
            lmax = self.lmax
        elif lmax > self.lmax:
            print("Warning: Cannot set lmax greater than the ", self.lmax)
            lmax = self.lmax

        for i in range(len(self.coefs)):
            ax1 = fig.add_subplot(gs[i + shift, 0])
            ax2 = fig.add_subplot(gs[i + shift, 1])
            coef = self.coefs[i]
            grid = coef.expand(lmax=lmax)
            grid.plot3d(0, 0, title=f"{self.ds[i]:6.3f}", show=False, ax=ax1)
            grid.plot(
                show=False,
                ax=ax2,
                tick_interval=[120, 90],
                tick_labelsize=14,
                axes_labelsize=16,
            )
            ax2.set_xlim([1, 359])

        if figname is None:
            plt.show()
        else:
            plt.savefig(figname)

    def plot_real_image(self, id=0):
        """
        Plot the real molecular contacts in the crystal
        """
        return self.xtal.show_mol_cluster(id, factor=self.factor, max_d=self.max_d, ignore_E=False, plot=False)

    def align(self, M=6):
        """
        Align spherical image in a way that three most important contributions
        are parallel to the equatorial plane. Experimental stage now!

        Args:
            M: number of power in quasi random sampling
        """
        self.coefs[0]
        angles = get_alignment(self.pts[0])
        self.coefs[0] = self.coefs[0].rotate(angles[0], angles[1], angles[2])

        for i in range(1, len(self.coefs)):
            coef1 = self.coefs[i]
            d, angles = correlation_go(self.coefs[0], coef1, M=M)
            self.coefs[i] = coef1.rotate(angles[0], angles[1], angles[2])
            self.ds[i] = d

    def rotate(self, alpha=0, beta=0, gamma=0):
        """
        uniformly rotate the coefs

        Args:
            alpha: rotation in degrees
            beta: rotation in degress
            gamma: rotation in degress
        """
        for i in range(len(self.coefs)):
            self.coefs[i] = self.coefs[i].rotate(alpha, beta, gamma)

    def get_similarity(self, sph2, M=6, cutoff=0.95):
        """
        Compute the similarity matrix between two sphs

        Args:
            sph2: the 2nd sph class
            M: number of power in quasi random sampling
            cutoff: cutoff similarity to terminate search early
        """
        S = np.zeros([len(self.coefs), len(sph2.coefs)])
        for i in range(len(self.coefs)):
            coef1 = self.coefs[i]
            for j in range(len(sph2.coefs)):
                coef2 = sph2.coefs[j]
                d, _ = correlation_go(coef1, coef2, M=M, d_cut=cutoff)
                S[i, j] = d
        return S


class orientation_order:
    """
    Computes the Steinhardt orientation order parameters

    Args:
        xtal: pyxtal structure
        max_d: maximum intermolecular distances
        lmax: maximum bandwidth for spherical harmonic expansion
    """

    def __init__(self, xtal, max_CN=14):
        self.xtal = xtal
        self.max_CN = max_CN
        self.dists = self.get_neighbors()

    def get_neighbors(self):
        """
        get neighboring molecules

        Returns:
            pts: [N, 3] array, (theta, phi, eng)
        """
        pts = []
        for i, site in enumerate(self.xtal.mol_sites):
            _, neighs, comps, _, engs = self.xtal.get_neighboring_molecules(i)
            xyz, _ = site._get_coords_and_species(absolute=True, first=True)
            center = site.molecule.get_center(xyz)
            coords = np.zeros([len(neighs), 3])
            # print(len(neighs))
            if len(neighs) > self.max_CN:
                neighs = neighs[: self.max_CN]
            for _i, xyz in enumerate(neighs):
                coords[_i, :] = self.xtal.molecules[comps[_i]].get_center(xyz) - center
            pts.append(coords)
        return pts

    def get_parameters(self, ls=None):
        """
        Computes
        Args:
            center: center xyz coordinate
            neighbors: a list of neighboring xyz coordinates
            weights: a list of weights for each neighbor
        Returns:
             q: numpy array(complex128), the complex vector qlm normalized
                by the number of nearest neighbors
        """
        if ls is None:
            ls = [4, 6]
        qs = []
        for dist in self.dists:
            for l in ls:
                (4 * np.pi) / (2 * l + 1)
                qlms = _qlm(dist, l)
                dot = float(np.sum(qlms * np.conjugate(qlms)))
                qs.append(np.sqrt((4 * np.pi) / (2 * l + 1) * dot))

        return qs


if __name__ == "__main__":
    import importlib.resources

    from pyxtal import pyxtal

    with importlib.resources.as_file(importlib.resources.files("pyxtal") / "database" / "cifs") as path:
        cif_path = path
    c1 = pyxtal(molecular=True)
    for name in ["benzene", "resorcinol", "aspirin", "naphthalene"]:
        c1.from_seed(seed=str(cif_path / f"{name}.cif"), molecules=[name])
        for model in ["contact", "molecule"]:
            print(name, model)
            sph = spherical_image(c1, model=model, lmax=18)
            sph.align()
            sph.plot_sph_images(figname=name + "-" + model + ".png")

    # for name in ['BENZEN', 'ACSALA', 'RESORA']:
    #    c1.from_CSD(name)
    #    for model in ['contact', 'molecule']:
    #        sph = spherical_image(c1, model=model, lmax=18)
    #        sph.align()
    #        sph.plot_sph_images(figname=name+'-'+model+'.png', molecule=True)



import json
import os.path as op

from pymatgen.core.structure import Molecule


class Collection:
    """Collection of molecular data.
    Used for obtaining pymatgen objects from a small database file.

    Example of use:

    >>> from pyxtal.database.collection import Collection
    >>> test=Collection('molecules')
    >>> test['H2O']
    Molecule Summary
    Site: O (0.0000, 0.0000, 0.0000)
    Site: H (0.2774, 0.8929, 0.2544)
    Site: H (0.6068, -0.2383, -0.7169)
    >>> list(test)
    ['C60', 'H2O', 'CH4', 'NH3', 'benzene', 'naphthalene', 'anthracene', 'tetracene', 'pentacene', 'coumarin', 'resorcinol', 'benzamide', 'aspirin', 'ddt', 'lindane', 'glycine', 'glucose', 'ROY']

    Args:
        name: the type of collection to get. Defaults to "molecules"
    """

    def __init__(self, name="molecules"):
        """Create a collection lazily.

        Will read data from json file when needed.

        A collection can be iterated over to get the Atoms objects and indexed
        with names to get individual members.

        Attributes:

        name: str
            Name of collection.
        data: object
            Pymetgen molecule object
        filename: str
            Location of json file.
        """

        self.name = name
        self._data = {}
        self.filename = op.join(op.dirname(__file__), name + ".json")
        with open(self.filename) as f:
            self.content = json.load(f)

    def __getitem__(self, name):
        self._read(name)
        if len(self._data) == 0:
            names = ""
            for dct in self.content:
                names += dct["name"] + ", "
            msg = name + " is not supported\n"
            msg += "Available molecules are:\n"
            msg += names
            raise NameError(msg)
        else:
            return self._data

    def __iter__(self):
        for dct in self.content:
            yield dct["name"]

    def _read(self, name):
        if self.name == "molecules":
            """
            read the data by name and convert it to pymatgen format
            """
            for dct in self.content:
                if dct["name"].lower() == name.lower():
                    pos = dct["xyz"]
                    symbols = dct["elements"]
                    self._data = Molecule(symbols, pos)
        elif self.name == "clusters":
            for dct in self.content:
                if dct["name"] == int(name):
                    self._data = dct

    def show_names(self):
        names = []
        for dct in self.content:
            names.append(dct["name"])
        print(names)



import os
from pyxtal.db import database
from pyxtal.interface.charmm import CHARMM
from ost.parameters import ForceFieldParameters

# Get pyxtal
db = database('pyxtal/database/test.db')
xtal = db.get_pyxtal('ACSALA') # missing data??
smiles = [mol.smile for mol in xtal.molecules]

# setup simulation
for i, method in enumerate(['am1bcc']):
    print('\n', method)
    dir_name = 'test1'
    if not os.path.exists(dir_name): os.mkdir(dir_name)
    params = ForceFieldParameters(smiles, 'gaff')
    parameters = params.params_init.copy()

    os.chdir(dir_name)
    xtal.to_file('init.cif')
    atoms = xtal.get_forcefield('gaff', code='charmm', chargemethod=method, parameters=parameters)
    atom_info = atoms.get_atom_info()
    calc = CHARMM(xtal, prefix='charmm', atom_info=atom_info)
    #print(calc.structure.lattice)
    calc.run() #clean=False)
    print(calc.structure.energy)
    print(calc.structure.lattice)
    calc.structure.to_file('opt.cif')



# python -m unittest pyxtal/test_all.py
import unittest
from pyxtal.lego.builder import mof_builder
from pyxtal import pyxtal
import random
import numpy as np

xtal = pyxtal()
xtal.from_spg_wps_rep(194, ['2c', '2b'], [2.46, 6.70])
cif_file = xtal.to_pymatgen()

builder1 = mof_builder(['C'], [1], verbose=False)
builder1.set_descriptor_calculator(mykwargs={'rcut': 1.9})
builder1.set_reference_enviroments(cif_file)
builder1.set_criteria(CN={'C': [3]})

xtal.from_spg_wps_rep(92, ['4a', '8b'], [5.085, 7.099, 0.294, 0.094, 0.241, 0.826], ['Si', 'O'])
cif_file = xtal.to_pymatgen()
builder2 = mof_builder(['Si', 'O'], [1, 2], verbose=False)
builder2.set_descriptor_calculator(mykwargs={'rcut': 2.4})
builder2.set_reference_enviroments(cif_file)
builder2.set_criteria(CN={'Si': [4], 'O': [2]}, exclude_ii=True)

class TestBuilder(unittest.TestCase):
    #def test_gen_xtal(self):
    #    print("test_gen_xtal")
    #    random.seed(0)
    #    np.random.seed(0)
    #    spg, wps = 191, [['6j', '6j', '6k']]
    #    xtal, sim = builder1.generate_xtal(spg, wps, 10,
    #                                       N_max = 1,
    #                                       add_db = False,
    #                                       random_state = 2)
    #    assert sim < 1e-2

    def test_opt_xtal(self):
        print("test_opt_xtal")
        spg, wps = 179, ['6a', '6a', '6a', '6a']
        for x in [
                    [11.111, 2.640, 0.290, 0.751, 0.421, 0.872],
                    [ 7.952, 2.606, 0.592, 0.926, 0.608, 0.307],
        ]:
            xtal = pyxtal()
            xtal.from_spg_wps_rep(spg, wps, x, ['C']*len(wps))
            xtal, sim, _ = builder1.optimize_xtal(xtal, add_db=False)
            #print(xtal.get_1d_rep_x())
            assert sim < 1e-2

    def test_opt_xtal2(self):
        print("test sio2")
        spg, wps = 92, ['4a', '8b']
        for x in [[5.0, 7.1, 0.3, 0.1, 0.2, 0.8]]:
            xtal = pyxtal()
            xtal.from_spg_wps_rep(spg, wps, x, ['Si', 'O'])
            xtal, sim, _ = builder2.optimize_xtal(xtal, add_db=False)
            assert sim < 1e-2

    def test_opt_xtals(self):
        print("test_opt_xtals")
        spg, wps = 179, ['6a', '6a', '6a', '6a']
        xtals = []
        for x in [
                    [11.111, 2.640, 0.290, 0.751, 0.421, 0.872],
                    [ 7.952, 2.606, 0.592, 0.926, 0.608, 0.307],
                 ]:
            xtal = pyxtal()
            xtal.from_spg_wps_rep(spg, wps, x, ['C']*len(wps))
            xtals.append(xtal)
        xtals = builder1.optimize_xtals(xtals, add_db=False)
        assert len(xtals) == 2
        #builder.optimize_xtals(xtals, opt_type='local', ncpu=2)

    def test_diamond(self):
        print("test_diamond")
        xtal = pyxtal()
        xtal.from_spg_wps_rep(227, ['8a'], [3.6], ['C'])
        xtals = xtal.subgroup(group_type='t+k')[7:]
        print("\nFirst split", xtal.group.number, len(xtals))
        for i, xtal1 in enumerate(xtals):
            xtal1s = xtal1.subgroup(group_type='t+k', max_cell=4)
            print("\n2nd split", i, xtal1.group.number, len(xtal1s))
            for _xtal1 in xtal1s[:3]:
                _xtal1.optimize_lattice(standard=True)
                final = _xtal1.subgroup_once(group_type='t+k')
                if final is not None:
                    #header = "{:3d} => ".format(xtal.group.number)
                    #header += "{:3d} => ".format(xtal1.group.number)
                    #header += "{:3d} => ".format(_xtal1.group.number)
                    #header += "{:3d} ".format(final.group.number)
                    #print(final.get_xtal_string(header=header))
                    builder1.optimize_xtal(final, add_db=False)

if __name__ == "__main__":
    unittest.main()



import os

import numpy as np

from pyxtal.database.element import Element


class Tol_matrix:
    """
    Class for variable distance tolerance checking. Used within random_crystal
    and molecular_crystal to verify whether atoms are too close. Stores a matrix
    of atom-atom pair tolerances. Note that the matrix's indices correspond to
    atomic numbers, with the 0th entries being 0 (there is no atomic number 0).

    Args:
        prototype: a string representing the type of radii to use
            (`atomic`, `molecular`, `vdW` or `metallic`)
        factor: a float to scale the distances by.
        tuples: a list or tuple of tuples, which define custom tolerance values.
            Each tuple should be of the form (specie1, specie2, value), where
            value is the tolerance in Angstroms, and specie1 and specie2 can be
            strings, integers, Element objects, or pymatgen Specie objects.
            Custom values may also be set using set_tol
    """

    def __init__(self, *tuples, prototype="atomic", factor=1.0):
        f = factor
        self.prototype = prototype
        if prototype == "atomic":
            f *= 0.5
            attrindex = 5
            self.radius_type = "covalent"
        elif prototype == "molecular":
            attrindex = 5
            self.radius_type = "covalent"
            f *= 1.2
        elif prototype == "metallic":
            attrindex = 7
            self.radius_type = "metallic"
            f *= 0.5
        elif prototype == "vdW":
            attrindex = 6
            self.radius_type = "vdW"
        else:
            self.radius_type = "N/A"

        self.f = f
        H = Element("H")
        m = [[0.0] * (len(H.elements_list) + 1)]
        for tup1 in H.elements_list:
            m.append([0.0])
            for tup2 in H.elements_list:
                # Get the appropriate atomic radii
                val1 = tup1[5] if tup1[attrindex] is None else tup1[attrindex]
                # FIXME this is suspicious but matches exactly the original code
                val2 = (tup1[5] if tup2[5] is not None else None) if tup2[attrindex] is None else tup2[attrindex]

                if val1 is not None and val2 is not None:
                    m[-1].append(f * (val1 + val2))
                else:
                    # If no radius is found for either atom, set tolerance to None
                    m[-1].append(None)
        self.matrix = np.array(m)  # A symmetric matrix between specie pairs
        self.custom_values = []  # A list of tuples storing customized pair tolerances

        try:
            for tup in tuples:
                self.set_tol(*tup)
        except Exception as err:
            msg = "Error: Cannot not set custom tolerance value(s).\n"
            msg += "All entries should be entered using the following form:\n"
            msg += "(specie1, specie2, value), where the value is in Angstrom."
            raise RuntimeError(msg) from err

        self.radius_list = []
        for i in range(len(self.matrix)):
            if i == 0:
                continue
            x = self.get_tol(i, i)
            self.radius_list.append(x)

    def get_tol(self, specie1, specie2):
        """
        Returns the tolerance between two species.

        Args:
            specie1/2: atomic number (int or float), name (str), symbol (str),
                an Element object, or a pymatgen Specie object

        Returns:
            the tolerance between the provided pair of atomic species
        """
        if self.prototype == "single value":
            return self.matrix[0][0]
        index1 = Element.number_from_specie(specie1)
        index2 = Element.number_from_specie(specie2)
        if index1 is not None and index2 is not None:
            return self.matrix[index1][index2]
        else:
            return None

    def set_tol(self, specie1, specie2, value):
        """
        Sets the distance tolerance between two species.

        Args:
            specie1/2: atomic number (int or float), name (str), symbol (str),
                an Element object, or a pymatgen Specie object
            value:
                the tolerance (in Angstroms) to set to
        """
        index1 = Element.number_from_specie(specie1)
        index2 = Element.number_from_specie(specie2)
        if index1 is None or index2 is None:
            return
        self.matrix[index1][index2] = float(value)
        if index1 != index2:
            self.matrix[index2][index1] = float(value)
        if (index1, index2) not in self.custom_values and (
            index2,
            index1,
        ) not in self.custom_values:
            larger = max(index1, index2)
            smaller = min(index1, index2)
            self.custom_values.append((smaller, larger))

    @classmethod
    def from_matrix(self, matrix, prototype="atomic", factor=1.0, begin_with=0):
        """
        Given a tolerance matrix, returns a Tol_matrix object. Matrix indices
        correspond to the atomic number (with 0 pointing to Hydrogen by default).
        For atoms with atomic numbers not included in the matrix, the default
        value (specified by prototype) will be used, up to element 96. Note that
        if the matrix is asymmetric, only the value below the diagonal will be used.

        Args:
            matrix: a 2D matrix or list of tolerances between atomic species pairs. The
                indices correspond to atomic species (see begin_with variable description)
            prototype: a string representing the type of radii to use
                ("atomic", "molecular")
            factor: a float to scale the distances by. A smaller value means a smaller
                tolerance for distance checking
            begin_with: the index which points to Hydrogen within the matrix. Default 0

        Returns:
            a Tol_matrix object
        """
        np.array(matrix)
        tups = []
        for i, row in enumerate(matrix):
            for j, _value in enumerate(row):
                if j > i:
                    continue
                tups.append((i + 1 - begin_with, j + 1 - begin_with, matrix[i][j]))
        return Tol_matrix(*tups, prototype=prototype, factor=factor)

    @classmethod
    def from_radii(self, radius_list, prototype="atomic", factor=1.0, begin_with=0):
        """
        Given a list of atomic radii, returns a Tol_matrix object. For atom-atom pairs, uses
        the average radii of the two species as the tolerance value. For atoms with atomic
        numbers not in the radius list, the default value (specified by prototype) will be
        used, up to element 96.

        Args:
            radius_list: a list of atomic radii (in Angstroms), beginning with Hydrogen
            prototype: a string representing the type of radii to use
                ("atomic", "molecular")
            factor: a float to scale the distances by. A smaller value means a smaller
                tolerance for distance checking
            begin_with: the index which points to Hydrogen within the list. Default 0

        Returns:
            a Tol_matrix object
        """
        tups = []
        f = factor * 0.5
        for i, r1 in enumerate(radius_list):
            for j, r2 in enumerate(radius_list):
                if j > i:
                    continue
                tups.append((i + 1 - begin_with, j + 1 - begin_with, f * (r1 + r2)))
        return Tol_matrix(*tups, prototype=prototype, factor=factor)

    @classmethod
    def from_single_value(self, value):
        """
        Creates a Tol_matrix which only has a single tolerance value. Using
        `get_tol` will always return the same value.

        Args:
            value: the tolerance value to use

        Returns:
            a Tol_matrix object
        """
        tm = Tol_matrix()
        tm.prototype = "single value"
        tm.matrix = np.array([[value]])
        tm.custom_values = [(1, 1)]
        tm.radius_type = "N/A"
        return tm

    def __getitem__(self, index):
        Element.number_from_specie(index)
        return self.matrix[index]

    def __str__(self):
        s = "\n--Tol_matrix class object--"
        s += "\nPrototype: " + str(self.prototype)
        s += "\nAtomic radius type: " + str(self.radius_type)
        s += "\nRadius scaling factor: " + str(self.f)
        if self.prototype == "single value":
            s += "\nCustom tolerance value: " + str(self.matrix([0][0]))
        else:
            if self.custom_values == []:
                s += "\nCustom tolerance values: None"
            else:
                s += "\nCustom tolerance values:"
                for tup in self.custom_values:
                    name1 = str(Element(tup[0]).short_name)
                    name2 = str(Element(tup[1]).short_name)
                    s += f"\n{name1:s}-{name2:s}: {self.get_tol(tup[0], tup[1]):6.3f}"
        return s

    def to_file(self, filename=None):
        """
        Creates a file with the given filename.

        Args:
            filename: the file path

        Returns:
            Nothing. Creates a file at the specified path
        """
        if filename is None:
            filename = "custom_tol_matrix"
        # Check if filename already exists
        # If it does, add a new number to end of filename
        if os.path.exists(filename):
            i = 1
            while True:
                outdir = filename + "_" + str(i)
                if not os.path.exists(outdir):
                    break
                i += 1
                if i > 10000:
                    return "Cannot create file: too many files already created."
        else:
            outdir = filename
        try:
            np.save(filename, [self])
            return "Output file to " + outdir + ".npy"
        except:
            return "Error: Could not save Tol_matrix to file."

    @classmethod
    def from_file(self, filename):
        try:
            tm = np.load(filename)[0]
            if type(tm) == Tol_matrix:
                return tm
            else:
                raise RuntimeError("invalid file for Tol_matrix: ", filename)
        except Exception as err:
            raise RuntimeError("Could not load Tol_matrix from file: ", filename) from err


if __name__ == "__main__":
    from pyxtal.molecule import pyxtal_molecule

    for p in ["atomic", "molecular", "vdW", "metallic"]:
        tm = Tol_matrix(prototype=p)
        print(p, tm.get_tol("C", "H"))
        m = pyxtal_molecule("aspirin", tm=tm)
        print(m.tols_matrix)



"""
Module for generating atomic crystals
"""

# Standard Libraries
from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np

from pyxtal.database.element import Element
from pyxtal.lattice import Lattice
from pyxtal.msg import Comp_CompatibilityError, VolumeError

# PyXtal imports #avoid *
from pyxtal.symmetry import Group, choose_wyckoff
from pyxtal.tolerance import Tol_matrix
from pyxtal.wyckoff_site import atom_site

if TYPE_CHECKING:
    from numpy.random import Generator


# Define functions
# ------------------------------
class random_crystal:
    """
    Class for storing and generating atomic crystals based on symmetry
    constraints. Given a spacegroup, list of atomic symbols, the stoichiometry,
    and a volume factor, generates a random crystal consistent with the
    spacegroup's symmetry.

    Args:
        dim: dimenion (0, 1, 2, 3)
        group: the group number (1-56, 1-75, 1-80, 1-230)
        species: a list of atomic symbols for each ion type, e.g., `["Ti", "O"]`
        numIons: a list of the number of each type of atom within the
            primitive cell (NOT the conventional cell), e.g., `[4, 2]`
        factor (optional): volume factor used to generate the crystal
        sites (optional): pre-assigned wyckoff sites (e.g., `[["4a"], ["2b"]]`)
        lattice (optional): `Lattice <pyxtal.lattice.Lattice.html>`_ object to
            define the unit cell
        tm (optional): `Tol_matrix <pyxtal.tolerance.Tol_matrix.html>`_ object
            to define the distances
    """

    def __init__(
        self,
        dim=3,
        group=227,
        species=None,
        numIons=8,
        factor=1.1,
        thickness=None,
        area=None,
        lattice=None,
        sites=None,
        conventional=True,
        tm=None,
        use_hall=False,
        random_state: int | None | Generator = None,
    ):
        # Initialize
        self.rng = np.random.default_rng(random_state)
        if species is None:
            species = ["C"]

        self.source = "Random"
        self.valid = False
        self.factor = factor
        self.min_density = 0.75

        # Dimesion
        self.dim = dim
        self.area = area  # Cross-section area for 1D
        self.thickness = thickness  # Thickness of 2D slab
        self.lattice0 = lattice

        # The periodic boundary condition
        if dim == 3:
            self.PBC = [1, 1, 1]
        elif dim == 2:
            self.PBC = [1, 1, 0]
        elif dim == 1:
            self.PBC = [0, 0, 1]
        elif dim == 0:
            self.PBC = [0, 0, 0]

        # Symmetry group
        if isinstance(group, Group):
            self.group = group
        else:
            self.group = Group(group, dim=self.dim, use_hall=use_hall)
        self.number = self.group.number
        self.symbol = self.group.symbol

        # Composition
        numIons = np.array(numIons)
        mul = self.group.cellsize() if not conventional else 1
        self.numIons = numIons * mul
        self.species = species

        # Tolerance matrix
        if tm is None:
            self.tol_matrix = Tol_matrix(prototype="atomic")
        elif isinstance(tm, Tol_matrix):
            self.tol_matrix = tm
        else:
            self.tol_matrix = Tol_matrix(prototype=tm)

        # Wyckoff sites
        self.set_sites(sites)

        # Lattice and coordinates
        compat, self.degrees = self.group.check_compatible(self.numIons)
        if not compat:
            self.valid = False
            msg = "Compoisition " + str(self.numIons)
            msg += " not compatible with symmetry "
            msg += str(self.group.number)
            raise Comp_CompatibilityError(msg)

        self.set_elemental_volumes()
        self.set_crystal()

    def __str__(self):
        if self.valid:
            s = f"------Crystal from {self.source:s}------"
            s += f"\nDimension: {self.dim}"
            s += f"\nGroup: {self.symbol} ({self.number})"
            s += f"\n{self.lattice}"
            s += "\nWyckoff sites:"
            for wyc in self.atom_sites:
                s += f"\n\t{wyc}"
        else:
            s = "\nStructure not available."
        return s

    def __repr__(self):
        return str(self)

    def set_sites(self, sites):
        """
        initialize Wyckoff sites
        Update 2023/09, track the wp index instead of letters to
        avoid many inquiries
        Args:
            sites: list
        """
        # Symmetry sites
        self.sites = {}
        for i, specie in enumerate(self.species):
            if sites is not None and sites[i] is not None and len(sites[i]) > 0:
                self.sites[specie] = []
                self._check_consistency(sites[i], self.numIons[i])

                if isinstance(sites[i], dict):
                    for item in sites[i].items():
                        # keep the record of wp index
                        id = self.group.get_index_by_letter(item[0])
                        self.sites[specie].append((id, item[1]))
                elif isinstance(sites[i], (list, tuple)):
                    for site in sites[i]:
                        if type(site) is tuple:
                            (letter, x, y, z) = site
                            id = self.group.get_index_by_letter(letter)
                            self.sites[specie].append((id, (x, y, z)))
                        else:
                            id = self.group.get_index_by_letter(site)
                            self.sites[specie].append(id)
            else:
                self.sites[specie] = None

    def set_elemental_volumes(self):
        """
        set up the radii for each specie
        """
        self.elemental_volumes = []
        for specie in self.species:
            sp = Element(specie)
            vol1, vol2 = sp.covalent_radius**3, sp.vdw_radius**3
            self.elemental_volumes.append([4 / 3 * np.pi * vol1, 4 / 3 * np.pi * vol2])

    def set_volume(self):
        """
        Estimates the volume of a unit cell based on the number/types of ions.
        Assumes each atom takes up a sphere with radius equal to its covalent
        bond radius.
        0.50 A -> 0.52 A^3
        0.62 A -> 1.00 A^3
        0.75 A -> 1.76 A^3

        Returns:
            a float value for the estimated volume
        """
        vmin_array = np.array([v[0] for v in self.elemental_volumes])
        vmax_array = np.array([v[1] for v in self.elemental_volumes])

        random_volumes = self.rng.uniform(vmin_array, vmax_array, size=len(self.numIons))
        volume = np.sum(np.array(self.numIons) * random_volumes)

        self.volume = self.factor * volume

        # make sure the volume is not too small
        if self.volume / sum(self.numIons) < self.min_density:
            self.volume = sum(self.numIons) * self.min_density

    def set_lattice(self, lattice):
        """
        Generate the initial lattice
        """
        if lattice is not None:
            # Use the provided lattice
            self.lattice = lattice
            self.volume = lattice.volume
            # Make sure the custom lattice PBC axes are correct.
            if lattice.PBC != self.PBC:
                self.lattice.PBC = self.PBC
        else:
            # Determine the unique axis
            if self.dim == 2:
                unique_axis = "c" if self.number in range(3, 8) else "a"
            elif self.dim == 1:
                unique_axis = "a" if self.number in range(3, 8) else "c"
            else:
                unique_axis = "c"

            # Generate a Lattice instance
            good_lattice = False
            for _cycle in range(10):
                try:
                    self.lattice = Lattice(
                        self.group.lattice_type,
                        self.volume,
                        PBC=self.PBC,
                        unique_axis=unique_axis,
                        thickness=self.thickness,
                        area=self.area,
                        random_state=self.rng,
                    )
                    good_lattice = True
                    break
                except VolumeError:
                    self.volume *= 1.1
                    warn(f"Warning: increase the volume by 1.1 times: {self.volume:.2f}")

            if not good_lattice:
                msg = f"Volume estimation {self.volume:.2f} is very bad with the given composition {self.numIons}"
                raise RuntimeError(msg)

    def set_crystal(self):
        """
        The main code to generate a random atomic crystal.
        If successful, `self.valid` is True
        """
        self.numattempts = 0
        if not self.degrees:
            self.lattice_attempts = 5
            self.coord_attempts = 5
        else:
            self.lattice_attempts = 40
            self.coord_attempts = 10

        # QZ: Maybe we no longer need this tag
        # if not self.lattice.allow_volume_reset:
        #    self.lattice_attempts = 1

        for cycle1 in range(self.lattice_attempts):
            self.set_volume()
            self.set_lattice(self.lattice0)
            # print("VOLUME", cycle1, self.volume)
            self.cycle1 = cycle1
            for cycle2 in range(self.coord_attempts):
                self.cycle2 = cycle2
                output = self._set_coords()

                if output:
                    self.atom_sites = output
                    break
            if self.valid:
                return
            else:
                self.lattice.reset_matrix()

        return

    def _set_coords(self):
        """
        generate coordinates for random crystal
        """
        wyks = []
        cell = self.lattice.matrix
        # generate coordinates for each ion type in turn
        for numIon, specie in zip(self.numIons, self.species):
            output = self._set_ion_wyckoffs(numIon, specie, cell, wyks)
            if output is not None:
                wyks.extend(output)
            else:
                # correct multiplicity not achieved exit and start over
                return None

        # If numIon_added correct for all specie return structure
        self.valid = True
        return wyks

    def _set_ion_wyckoffs(self, numIon, specie, cell, wyks):
        """
        generates a set of wyckoff positions to accomodate a given number
        of ions

        Args:
            numIon: Number of ions to accomodate
            specie: Type of species being placed on wyckoff site
            cellx: Matrix of lattice vectors
            wyks: current wyckoff sites

        Returns:
            Sucess:
                wyckoff_sites_tmp: list of wyckoff sites for valid sites
            Failue:
                None

        """
        numIon_added = 0
        tol = self.tol_matrix.get_tol(specie, specie)
        wyckoff_sites_tmp = []

        # Now we start to add the specie to the wyckoff position
        sites_list = deepcopy(self.sites[specie])  # the list of Wyckoff site
        if sites_list is not None:
            wyckoff_attempts = max(len(sites_list) * 2, 10)
        else:
            # the minimum numattempts is to put all atoms to the general WPs
            min_wyckoffs = int(numIon / len(self.group.wyckoffs_organized[0][0]))
            wyckoff_attempts = max(2 * min_wyckoffs, 10)

        cycle = 0
        while cycle < wyckoff_attempts:
            # Choose a random WP for given multiplicity: 2a, 2b
            site = sites_list[0] if sites_list is not None and len(sites_list) > 0 else None

            new_site = None
            if type(site) is tuple:  # site with coordinates
                # print(site)
                (index, xyz) = site
                wp = self.group[index]
                new_site = atom_site(wp, xyz, specie)
            else:
                if site is not None:
                    wp = self.group[site]
                    pt = self.lattice.generate_point()
                    # avoid using the merge function
                    if len(wp.short_distances(pt, cell, tol)) > 0:
                        # print('bad pt', pt, wp.short_distances(pt, cell, tol))
                        cycle += 1
                        continue
                    # update pt
                    pt = wp.project(pt, cell, self.PBC)
                    # print('good', pt, tol, len(wp.short_distances(pt, cell, tol)))
                else:
                    # generate wp
                    wp = choose_wyckoff(self.group, numIon - numIon_added, site, self.dim, self.rng)
                    if wp is not False:
                        # print(wp.letter)
                        # Generate a list of coords from ops
                        pt = self.lattice.generate_point()
                        pt, wp, _ = wp.merge(pt, cell, tol, group=self.group)
                # print('good pt', pt)
                if wp is not False:
                    # For pure planar structure
                    if self.dim == 2 and self.thickness is not None and self.thickness < 0.1:
                        pt[-1] = 0.5
                    new_site = atom_site(wp, pt, specie)

            # Check current WP against existing WP's
            if self.check_wp(wyckoff_sites_tmp, wyks, cell, new_site):
                if sites_list is not None and len(sites_list) > 0:
                    sites_list.pop(0)
                wyckoff_sites_tmp.append(new_site)
                numIon_added += new_site.multiplicity

                # Check if enough atoms have been added
                if numIon_added == numIon:
                    return wyckoff_sites_tmp

            cycle += 1
            self.numattempts += 1

        return None

    def check_wp(self, wyckoff_sites_tmp, wyks, cell, new_site):
        # Check current WP against existing WP's
        if new_site is None:
            return False

        return all(new_site.check_with_ws2(ws, cell, self.tol_matrix) for ws in wyckoff_sites_tmp + wyks)

    def _check_consistency(self, site, numIon):
        num = 0
        for s in site:
            if isinstance(s, dict):
                for key in s:
                    num += int(key[:-1])
            else:
                num += int(s[:-1])
        if numIon == num:
            return True
        else:
            diff = numIon - num
            if diff > 0:
                # check if compatible
                compat, self.degrees = self.group.check_compatible([diff])
                if compat:
                    return True
                else:
                    msg = (
                        f"\nfrom numIons: {numIon:d}"
                        f"\nfrom Wyckoff list: {num:d}"
                        f"\nThe number is incompatible with composition: {site}"
                    )
                raise ValueError(msg)

            msg = (
                f"\nfrom numIons: {numIon:d}"
                f"\nfrom Wyckoff list: {num:d}"
                f"\nThe requested number is greater than composition: {site}"
            )
            raise ValueError(msg)



#!/usr/bin/env  python

import os
from argparse import ArgumentParser

import numpy as np

from pyxtal import print_logo, pyxtal
from pyxtal.symmetry import get_symbol_and_number

if __name__ == "__main__":
    # -------------------------------- Options -------------------------
    parser = ArgumentParser()
    parser.add_argument(
        "-s",
        "--symmetry",
        dest="sg",
        metavar="sg",
        default=36,
        type=str,
        help="desired symmetry, number or string, e.g., 36, Pbca, Ih",
    )
    parser.add_argument(
        "-e",
        "--element",
        dest="element",
        default="Li",
        help="desired elements: e.g., Li",
        metavar="element",
    )
    parser.add_argument(
        "-n",
        "--numIons",
        dest="numIons",
        default=16,
        help="desired numbers of atoms: 16",
        metavar="numIons",
    )
    parser.add_argument(
        "-f",
        "--factor",
        dest="factor",
        default=1.0,
        type=float,
        help="volume factor: default 1.0",
        metavar="factor",
    )
    parser.add_argument(
        "-a",
        "--attempts",
        dest="attempts",
        default=1,
        type=int,
        help="number of crystals to generate: default 1",
        metavar="attempts",
    )
    parser.add_argument(
        "-o",
        "--outdir",
        dest="outdir",
        default="out",
        type=str,
        help="Directory for storing output cif files: default 'out'",
        metavar="outdir",
    )
    parser.add_argument(
        "-d",
        "--dimension",
        dest="dimension",
        metavar="dimension",
        default=3,
        type=int,
        help="desired dimension: (3, 2, 1, 0): default 3",
    )
    parser.add_argument(
        "-t",
        "--thickness",
        dest="thickness",
        metavar="thickness",
        default=None,
        type=float,
        help="Thickness of a 2D crystal, or area of a 1D crystal, None generates a value automatically: default None",
    )

    parser.add_argument(
        "-m",
        "--molecular",
        dest="molecular",
        action="store_true",
        default=False,
        help="molecular? default: False",
    )

    parser.add_argument(
        "-c",
        "--conventional",
        dest="conventional",
        action="store_true",
        default=False,
        help="conventional setting? default: False",
    )

    print_logo()
    options = parser.parse_args()
    sg = options.sg
    dimension = options.dimension
    if isinstance(sg, str) and sg.isnumeric():
        sg = int(sg)
    symbol, sg = get_symbol_and_number(sg, dimension)

    element = options.element
    number = options.numIons
    numIons = []
    if element.find(",") > 0:
        system = element.split(",")
        for x in number.split(","):
            numIons.append(int(x))
    else:
        system = [element]
        numIons = [int(number)]

    factor = options.factor
    if factor < 0:
        raise ValueError(f"Volume factor {factor:.2f} must be greater than 0.")

    attempts = options.attempts
    outdir = options.outdir
    dimension = options.dimension
    thickness = options.thickness
    molecular = options.molecular
    conventional = options.conventional

    if not os.path.exists(outdir):
        os.mkdir(outdir)

    for i in range(attempts):
        numIons0 = np.array(numIons)
        rand_crystal = pyxtal(molecular=molecular)
        if dimension == 3:
            rand_crystal.from_random(3, sg, system, numIons0, factor, conventional=conventional)
        elif dimension == 2:
            rand_crystal.from_random(2, sg, system, numIons0, factor, thickness, conventional=conventional)
        elif dimension == 1:
            rand_crystal.from_random(1, sg, system, numIons0, factor, thickness, conventional=conventional)
        if dimension == 0:
            rand_crystal.from_random(0, sg, system, numIons0, factor)
        # Output a cif or xyz file
        outpath = options.outdir + "/" + str(i) + ".cif" if dimension > 0 else options.outdir + "/" + str(i) + ".xyz"
        rand_crystal.to_file(filename=outpath)

        print(rand_crystal)



"""
A base class for global optimization including:

- WFS
- DFS
- QRS
"""
from __future__ import annotations
from multiprocessing import Pool

from concurrent.futures import TimeoutError
import signal

import logging
import os
from time import time
from typing import TYPE_CHECKING

import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher
from numpy.random import Generator

from pyxtal.molecule import find_rotor_from_smile, pyxtal_molecule
from pyxtal.representation import representation
from pyxtal.util import new_struc
from pyxtal.optimize.common import optimizer, randomizer
from pyxtal.optimize.common import optimizer_par, optimizer_single
from pyxtal.lattice import Lattice
from pyxtal.symmetry import Group

def setup_worker_logger(log_file):
    """
    Set up the logger for each worker process.
    """
    logging.getLogger().handlers.clear()
    logging.basicConfig(format="%(asctime)s| %(message)s",
                        filename=log_file,
                        level=logging.INFO)


# Update run_optimizer_with_timeout to accept a logger
def run_optimizer_with_timeout(args, logger):
    """
    Run the optimizer with a timeout.
    This function will be executed by each process.
    """
    def handler(signum, frame):
        raise TimeoutError("Optimization timed out")

    # Set the timeout signal
    cwd = os.getcwd()
    timeout = int(args[-1])
    #logger.info(f"Rank-{args[-2]} entering optimizer_with_timeout")
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(timeout)
    #logger.info(f"Rank-{args[-2]} after signal")

    try:
        if args[-2] > 0:
            logger.info(f"Rank-{args[-2]} running optimizer_par for PID {os.getpid()}")
        result = optimizer_par(*args[:-2])
        if args[-2] > 0:
            logger.info(f"Rank-{args[-2]} finished optimizer_par for PID {os.getpid()}")
        signal.alarm(0)  # Disable the alarm
        return result
    except TimeoutError:
        logger.info(f"Rank-{args[-2]} Process {os.getpid()} timed out after {timeout} seconds.")
        os.chdir(cwd)
        return None  # or some other placeholder for timeout results

# Update process_task to accept a logger
def process_task(args):
    logger = logging.getLogger()
    #logger.info(f"Rank {args[-2]} start process_task.")
    result = run_optimizer_with_timeout(args, logger)
    return result

class GlobalOptimize:
    """
    Base-class for all global optimization methods

    Args:
        smiles (str): smiles string
        workdir (str): path of working directory
        sg (int or list): space group number or list of spg numbers
        tag (string): job prefix
        ff_opt (bool): activate on the fly FF mode
        ff_style (str): automated force style (`gaff` or `openff`)
        ff_parameters (str or list): ff parameter xml file or list
        reference_file (str): path of reference xml data for FF training
        N_cpu (int): number of cpus for parallel calculation (default: `1`)
        cif (str): cif file name to store all structure information
        block: block mode
        num_block: list of blocks
        compositions: list of composition, (default is [1]*Num_mol)
        lattice (bool): whether or not supply the lattice
        torsions: list of torsion angle
        molecules (list): list of pyxtal_molecule objects
        sites (list): list of wp sites, e.g., [['4a']]
        use_hall (bool): whether or not use hall number (default: False)
        skip_ani (bool): whether or not use ani or not (default: True)
        eng_cutoff (float): the cutoff energy for FF training
        E_max (float): maximum energy defined as an invalid structure
        matcher : structurematcher from pymatgen
        early_quit: whether quit the program early when the target is found
        pre_opt: whether pre_optimize the structure or not
    """

    def __init__(
        self,
        smiles: str,
        workdir: str,
        sg: int | list[int],
        tag: str,
        info: dict[any, any] | None = None,
        ff_opt: bool = False,
        ff_style: str = "openff",
        ff_parameters: str = "parameters.xml",
        reference_file: str = "references.xml",
        ref_criteria: dict[any, any] | None = None,
        N_cpu: int = 1,
        cif: str | None = None,
        block: list[any] | None = None,
        num_block: list[any] | None = None,
        composition: list[any] | None = None,
        lattice: Lattice | None = None,
        torsions: list[any] | None = None,
        molecules: list[pyxtal_molecule] | None = None,
        sites: list[any] | None = None,
        use_hall: bool = False,
        skip_ani: bool = True,
        factor: float = 1.1,
        eng_cutoff: float = 5.0,
        E_max: float = 1e10,
        random_state=None,
        max_time: float | None = None,
        matcher: StructureMatcher | None = None,
        early_quit: bool = True,
        check_stable: bool = False,
        use_mpi: bool = False,
        pre_opt: bool = False,
    ):

        self.ncpu = N_cpu
        self.use_mpi = use_mpi
        if self.use_mpi:
            from mpi4py import MPI
            self.comm = MPI.COMM_WORLD
            self.rank = self.comm.Get_rank()
            self.size = self.comm.Get_size()
        else:
            self.rank = 0
            self.size = self.ncpu

        # General information
        if isinstance(random_state, Generator):
            self.random_state = random_state.spawn(1)[0]
        else:
            self.random_state = np.random.default_rng(random_state)

        # Molecular information
        self.smile = smiles
        self.smiles = self.smile.split(".")  # list
        self.torsions = torsions
        self.molecules = molecules
        self.block = block
        self.num_block = num_block
        self.composition = [
            1] * len(self.smiles) if composition is None else composition
        self.N_torsion = 0
        for smi, comp in zip(self.smiles, self.composition):
            self.N_torsion += len(find_rotor_from_smile(smi)
                                  ) * int(max([comp, 1]))

        # Crystal information
        self.pre_opt = pre_opt
        self.sg = [sg] if isinstance(sg, (int, np.int64)) else sg
        self.use_hall = use_hall
        self.factor = factor
        self.sites = sites
        if lattice is None:
            self.lattice = lattice
        elif isinstance(lattice, Lattice):
            self.lattice = lattice
        else:
            ltype = Group(self.sg[0]).lattice_type
            if len(lattice) == 6:
                self.lattice = Lattice.from_para(lattice, ltype=ltype)
            else:
                raise ValueError("input lattice is invalid", lattice)
        self.opt_lat = self.lattice is None
        self.ref_criteria = ref_criteria
        self.eng_cutoff = eng_cutoff

        # Generation and Optimization
        self.workdir = workdir
        os.makedirs(self.workdir, exist_ok=True)
        self.log_file = self.workdir + "/loginfo"
        if self.rank > 0: self.log_file += f"-{self.rank}"

        self.skip_ani = skip_ani
        # self.randomizer = randomizer
        # self.optimizer = optimizer
        self.check_stable = check_stable
        if not self.opt_lat:
            self.check_stable = False
        # setup timeout for each optimization call
        if max_time is None:
            if not self.skip_ani:
                max_time = 300.0
            elif self.check_stable:
                max_time = 300.0
            else:
                max_time = 60.0
        self.timeout = max_time * self.N_pop / self.ncpu

        self.ff_opt = ff_opt
        self.ff_style = ff_style

        # Setup logger
        logging.getLogger().handlers.clear()
        logging.basicConfig(format="%(asctime)s| %(message)s",
                            filename=self.log_file, level=logging.INFO)
        self.logging = logging

        if info is not None:
            self.atom_info = info
            self.parameters = None
            self.ff_opt = False
        else:
            self.ff_parameters = self.workdir + "/" + ff_parameters
            self.reference_file = self.workdir + "/" + reference_file
            # Only call ForceFieldParameters once
            # No need to broadcast self.parameters?
            # Just broadcast atom_info should be fine
            # parameters = None
            atom_info = None
            if self.rank == 0:
                from pyocse.parameters import ForceFieldParameters
                self.parameters = ForceFieldParameters(
                    self.smiles, style=ff_style,
                    f_coef=1.0,
                    s_coef=1.0,
                    ref_evaluator='mace',
                    ncpu=self.ncpu)

                # Preload two set for FF parameters 1 for opt and 2 for refinement
                if isinstance(self.ff_parameters, list):
                    assert len(self.ff_parameters) == 2
                    for para_file in self.ff_parameters:
                        if not os.path.exists(para_file):
                            raise RuntimeError("File not found", para_file)
                    params0, dic = self.parameters.load_parameters(
                        self.ff_parameters[0])
                    if "ff_style" in dic:
                        assert dic["ff_style"] == self.ff_style
                    # print(params0)
                    params1, dic = self.parameters.load_parameters(
                        self.ff_parameters[1])
                    if "ff_style" in dic:
                        assert dic["ff_style"] == self.ff_style
                    # print(params1)
                    atom_info = self._prepare_chm_info(params0, params1)
                else:
                    if os.path.exists(self.ff_parameters):
                        self.print("Preload the existing FF parameters from",
                                   self.ff_parameters)
                        params0, _ = self.parameters.load_parameters(
                            self.ff_parameters)
                    else:
                        self.print(
                            "No FF parameter file exists, using the default setting",
                            ff_style,
                        )
                        params0 = self.parameters.params_init.copy()
                        self.parameters.export_parameters(self.ff_parameters, params0)
                    atom_info = self._prepare_chm_info(
                        params0, suffix='pyxtal')

            if self.use_mpi:
                # self.parameters = self.comm.bcast(parameters, root=0)
                self.atom_info = self.comm.bcast(atom_info, root=0)
            else:
                self.atom_info = atom_info

        # Structure matcher
        if matcher is None:
            self.matcher = StructureMatcher(ltol=0.3, stol=0.3, angle_tol=5)
        else:
            self.matcher = matcher

        # I/O stuff
        self.early_quit = early_quit
        self.N_min_matches = 10  # The min_num_matches for early termination
        self.E_max = E_max
        self.tag = tag.lower()
        self.suffix = f"{self.workdir:s}/{self.name:s}-{self.ff_style:s}"
        if self.rank == 0:
            if cif is None:
                self.cif = self.suffix + '.cif'
            else:
                self.cif = self.suffix + cif

            with open(self.cif, "w") as f:
                f.writelines(str(self))
            self.matched_cif = self.suffix + "-matched.cif"
            # print(self)

            # Some neccessary trackers
            self.matches = []
            self.best_reps = []
            self.reps = []
            self.engs = []

    def print(self, *args, **kwargs):
        """Utility method to print only from rank 0."""
        if self.rank == 0:
            print(*args, **kwargs)

    def __str__(self):
        s = "\n-------Global Crystal Structure Prediction------"
        s += f"\nsmile     : {self.smile:s}"
        s += f"\nZprime    : {self.composition!s:s}"
        s += f"\nN_torsion : {self.N_torsion:d}"
        s += f"\nsg        : {self.sg!s:s}"
        s += f"\nncpu      : {self.size:d}"
        s += f"\ndiretory  : {self.workdir:s}"
        s += f"\nopt_lat   : {self.opt_lat!s:s}"
        s += f"\nusp_mpi   : {self.use_mpi!s:s}\n"
        if self.early_quit:
            s += f"Mode      : Production\n"
        else:
            s += f"Mode      : Sampling\n"
        s += f"cif       : {self.cif:s}\n"
        if self.ff_opt:
            s += "forcefield: On-the-fly\n"
        else:
            s += "forcefield: Predefined\n"

        if self.parameters is not None:
            s += f"ff_style  : {self.ff_style:s}\n"
            if isinstance(self.ff_parameters, list):
                for para in self.ff_parameters:
                    s += f"ff_params : {para:s}\n"
            else:
                s += f"ff_params : {self.ff_parameters:s}\n"
            s += f"references: {self.reference_file:s}\n"
            s += str(self.parameters)

        return s

    def __repr__(self):
        return str(self)

    def new_struc(self, xtal, xtals):
        return new_struc(xtal, xtals)

    def run(self, ref_pmg=None, ref_pxrd=None):
        """
        The main code to run Sampling

        Args:
            ref_pmg: reference pmg structure
            ref_pxrd: reference pxrd profile in 2D array

        Returns:
            success_rate or None
        """
        t0 = time()

        if ref_pmg is not None:
            ref_pmg.remove_species("H")
        self.ref_pmg = ref_pmg
        self.ref_pxrd = ref_pxrd

        if self.ncpu > 1:
            pool = Pool(processes=self.ncpu, initializer=setup_worker_logger, initargs=(self.log_file,))
        else:
            pool = None

        results = self._run(pool)

        if self.rank == 0:
            t = (time() - t0)/60
            strs = f"{self.name:s} {self.workdir} COMPLETED "
            strs += f"in {t:.1f} mins {self.N_struc:d} strucs."
            print(strs)

        if self.use_mpi:
            self.comm.Barrier()
        return results

    def select_xtals(self, ref_xtals, ids, N_max):
        """
        Select only unique structures
        """
        xtals = []
        for id in ids:
            (xtal, _) = ref_xtals[id]
            if xtal.energy <= self.E_max and self.new_struc(xtal, xtals):
                xtals.append(xtal)  # .to_ase(resort=False))
            if len(xtals) == N_max:
                break
        # xtals = [xtal.to_ase(resort=False) for xtal in xtals]
        return xtals

    def count_pxrd_match(self, xtals, matches):
        """
        Wrap up the matched PXRD results

        Args:
            xtals: list of (xtal, tag) tuples
            matches (list): list of XRD matches

        """
        gen = self.generation
        for i, match in enumerate(matches):
            if match > 0.85:
                (xtal, tag) = xtals[i]
                with open(self.matched_cif, "a+") as f:
                    e = xtal.energy / sum(xtal.numMols)
                    try:
                        label = self.tag + "-g" + str(gen) + "-p" + str(i)
                        label += f"-e{e:.3f}-{tag:s}-{match:4.2f}"
                    except:
                        print("Error in e, tag, match", e, tag, match)
                    f.writelines(xtal.to_file(header=label))
                    self.matches.append((gen, i, xtal, e, match, tag))

    def success_count(self, xtals, matches):
        """
        Wrap up the matched results and count success rate.

        Args:
            xtals: list of (xtal, tag) tuples
            matches (list): list of matches [True, False, ..]

        Return:
            success_rate
        """
        gen = self.generation
        for i, match in enumerate(matches):
            if match:
                (xtal, tag) = xtals[i]
                with open(self.matched_cif, "a+") as f:
                    res = self._print_match(xtal, self.ref_pmg)
                    e, d1, d2 = xtal.energy/sum(xtal.numMols), res[0], res[1]
                    try:
                        label = self.tag + "-g" + str(gen) + "-p" + str(i)
                        label += f"-e{e:.3f}-{tag:s}-{d1:4.2f}-{d2:4.2f}"
                    except:
                        print("Error in e, tag, d1, d2", e, tag, d1, d2)
                    f.writelines(xtal.to_file(header=label))
                    self.matches.append((gen, i, xtal, e, d1, d2, tag))

        success_rate = len(self.matches) / self.N_struc * 100
        gen_out = f"Success rate @ Gen {gen:3d}: {success_rate:7.4f}%"
        self.logging.info(gen_out)
        print(gen_out)

        return success_rate

    def early_termination(self, success_rate):
        """
        Check if the calculation can be terminated early.
        """
        if success_rate > 0:
            if self.early_quit:
                msg = f"Early termination since a match is found"
                print(msg)
                self.logging.info(msg)
                return True

            elif success_rate > 2.5 or len(self.matches) >= self.N_min_matches:
                msg = f"Early termination with a high success rate"
                print(msg)
                self.logging.info(msg)
                return True
        return False

    def update_ff_paramters(self, xtals, engs, N_added):
        """
        Update the ff parameters

        Args:
            xtals: a list of pyxtals
            engs: a list of energies
            N_added (int): the number of structures that have been added
        """

        N_max = min([int(self.N_pop * 0.6), 50])
        ids = np.argsort(engs)
        _xtals = self.select_xtals(xtals, ids, N_max)
        print("Select structures for FF optimization", len(_xtals))

        return self.ff_optimization(_xtals, N_added)

    def ff_optimization(self, xtals, N_added, N_min=50, dE=2.5, FMSE=2.5):
        """
        Optimize the current FF based on newly explored data

        Args:
            xtals: list of xtals from the current gen
            N_added (int): the number of structures that have been added
            N_min (int): minimum number of configurations to trigger the FF training
            dE (float): the cutoff energy value
            FMSE (float): the cutoff Force MSE value
        """
        from pyocse.utils import reset_lammps_cell
        from pyocse.parameters import compute_r2, get_lmp_efs

        numMols = [xtal.numMols for xtal in xtals]
        xtals = [xtal.to_ase(resort=False) for xtal in xtals]

        gen = self.generation
        # Initialize ff parameters and references
        params_opt, err_dict = self.parameters.load_parameters(
            self.ff_parameters)
        if os.path.exists(self.reference_file):
            ref_dics = self.parameters.load_references(self.reference_file)
            if self.ref_criteria is not None:
                ref_dics = self.parameters.clean_ref_dics(
                    ref_dics, self.ref_criteria)
                ref_dics = self.parameters.cut_references_by_error(
                    ref_dics, params_opt, dE=dE, FMSE=FMSE)
            # self.parameters.generate_report(ref_dics, params_opt)
        else:
            ref_dics = []

        # Add references
        print("Current number of reference structures", len(ref_dics))
        t0 = time()
        if len(ref_dics) > 100:  # no fit if ref_dics is large
            # Here we find the lowest engs and select only low-E struc
            ref_engs = [ref_dic["energy"] / ref_dic["replicate"]
                        for ref_dic in ref_dics]
            ref_e2 = np.array(ref_engs).min()
            print("Min Reference Energy", ref_e2)

            if len(err_dict) == 0:
                # update the offset if necessary
                _, params_opt = self.parameters.optimize_offset(
                    ref_dics, params_opt)
                results = self.parameters.evaluate_multi_references(
                    ref_dics, params_opt, 1000, 1000)
                (ff_values, ref_values, rmse_values, r2_values) = results
                err_dict = {"rmse_values": rmse_values}

            _ref_dics = []
            rmse_values = err_dict["rmse_values"]
            lmp_in = self.parameters.ff.get_lammps_in()
            self.parameters.ase_templates = {}
            self.lmp_dat = {}
            ff_engs, ff_fors, ff_strs = [], [], []
            rf_engs, rf_fors, rf_strs = [], [], []

            for numMol, xtal in zip(numMols, xtals):
                struc = reset_lammps_cell(xtal)
                lmp_struc, lmp_dat = self.parameters.get_lmp_input_from_structure(
                    struc, numMol, set_template=False)
                replicate = len(lmp_struc.atoms) / \
                    self.parameters.natoms_per_unit

                try:
                    # ; print('Debug KONTIQ', struc, e1)
                    e1, f1, s1 = get_lmp_efs(lmp_struc, lmp_in, lmp_dat)
                except:
                    e1 = self.E_max

                # filter very high energy structures
                if e1 < 1000:
                    struc.set_calculator(self.parameters.calculator)
                    e2 = struc.get_potential_energy()
                    e2 /= replicate
                    # Ignore very high energy structures
                    if e2 < ref_e2 + self.eng_cutoff:
                        e1 /= replicate
                        f2 = struc.get_forces()
                        s2 = struc.get_stress()
                        struc.set_calculator()
                        e_err = abs(e1 - e2 + params_opt[-1])
                        f_err = np.sqrt(
                            ((f1.flatten() - f2.flatten()) ** 2).mean())
                        s_err = np.sqrt(((s1 - s2) ** 2).mean())
                        e_check = e_err < 0.5 * rmse_values[0]
                        f_check = f_err < 1.0 * rmse_values[1]
                        s_check = s_err < 1.0 * rmse_values[2]
                        strs = f"Errors of csp structure in gen{gen:3d} "
                        strs += f"{e_err:.4f} {f_err:.4f} {s_err:.4f} "
                        strs += f"{e1:8.4f} {e2:8.4f}"
                        print(strs, e_check, f_check, s_check)

                        # avoid very unphysical structures
                        if e_err < 4.0 and f_err < 4.0:
                            ff_engs.append(e1 + params_opt[-1])
                            rf_engs.append(e2)
                            ff_fors.extend(f1.flatten())
                            rf_fors.extend(f2.flatten())
                            ff_strs.extend(s1.flatten())
                            rf_strs.extend(s2.flatten())

                            if False in [e_check, f_check, s_check]:
                                _ref_dic = {
                                    "structure": struc,
                                    "energy": e2 * replicate,
                                    "forces": f2,
                                    "stress": s2,
                                    "replicate": replicate,
                                    "options": [True, not f_check, True],
                                    "tag": "CSP",
                                    "numMols": numMol,
                                }
                                _ref_dics.append(_ref_dic)
                    else:
                        print("Ignore the structure due to high energy", e2)

            # QZ: Output FF performances MSE, R2 for the selected structures
            if len(_ref_dics) == 0:
                print("There is a serious problem in depositing high energy")
                raise ValueError("The program needs to stop here")

            if self.ref_criteria is not None:
                _ref_dics = self.parameters.clean_ref_dics(
                    _ref_dics, self.ref_criteria)

            # print("Added {:d} new reference structures into training".format(len(_ref_dics)))
            print("FF performances")
            ff_engs = np.array(ff_engs)
            rf_engs = np.array(rf_engs)
            ff_fors = np.array(ff_fors)
            rf_fors = np.array(rf_fors)
            ff_strs = np.array(ff_strs)
            rf_strs = np.array(rf_strs)
            r2_engs = compute_r2(ff_engs, rf_engs)
            r2_fors = compute_r2(ff_fors, rf_fors)
            r2_strs = compute_r2(ff_strs, rf_strs)
            mse_engs = np.sqrt(np.mean((ff_engs - rf_engs) ** 2))
            mse_fors = np.sqrt(np.mean((ff_fors - rf_fors) ** 2))
            mse_strs = np.sqrt(np.mean((ff_strs - rf_strs) ** 2))

            print(f"R2   {r2_engs:8.4f} {r2_fors:8.4f} {r2_strs:8.4f}")
            print(f"RMSE {mse_engs:8.4f} {mse_fors:8.4f} {mse_strs:8.4f}")
            # self.parameters.generate_report(_ref_dics, params_opt)
            # import sys; sys.exit()
        else:
            # reduce the number of structures to save some time
            N_selected = min([N_min, self.ncpu])
            print("Create the reference data by augmentation", N_selected)
            if len(xtals) >= N_selected:
                ids = self.random_state.choice(
                    list(range(len(xtals))), N_selected)
                xtals = [xtals[id] for id in ids]
                numMols = [numMols[id] for id in ids]

            _ref_dics = self.parameters.add_multi_references(
                xtals,
                numMols,
                augment=True,
                steps=20,  # 50,
                N_vibs=1,
                logfile="ase.log",
            )
            if len(ref_dics) == 0:
                _, params_opt = self.parameters.optimize_offset(
                    _ref_dics, params_opt)
                self.parameters.generate_report(ref_dics, params_opt)
                # import sys; sys.exit()

        ref_dics.extend(_ref_dics)
        print(
            f"Add {len(_ref_dics):d} references in {(time() - t0) / 60:.2f} min")
        self.parameters.export_references(ref_dics, self.reference_file)

        # Optimize ff parameters if we get enough number of configurations
        N_added += len(_ref_dics)
        if N_added < N_min:
            print("Do not update ff, the current number of configurations is", N_added)
        else:
            t0 = time()
            _, params_opt = self.parameters.optimize_offset(
                ref_dics, params_opt)

            for data in [
                (["bond", "angle", "proper"], 50),
                (["proper", "vdW", "charge"], 50),
                (["bond", "angle", "proper", "vdW", "charge"], 50),
            ]:
                (terms, steps) = data

                # Actual optimization
                opt_dict = self.parameters.get_opt_dict(
                    terms, None, params_opt)
                x, fun, values, it = self.parameters.optimize_global(
                    ref_dics, opt_dict, params_opt, obj="R2", t0=0.1, steps=25
                )

                params_opt = self.parameters.set_sub_parameters(
                    values, terms, params_opt)

                opt_dict = self.parameters.get_opt_dict(
                    terms, None, params_opt)

                x, fun, values, it = self.parameters.optimize_local(
                    ref_dics, opt_dict, params_opt, obj="R2", steps=steps)

                params_opt = self.parameters.set_sub_parameters(
                    values, terms, params_opt)
                _, params_opt = self.parameters.optimize_offset(
                    ref_dics, params_opt)

                # To add Early termination

            t = (time() - t0) / 60
            print(f"FF optimization {t:.2f} min ", fun)
            # Reset N_added to 0
            N_added = 0

        # Export FF performances
        if gen < 10:
            gen_prefix = "gen_00" + str(gen)
        elif gen < 100:
            gen_prefix = "gen_0" + str(gen)
        else:
            gen_prefix = "gen_" + str(gen)

        performance_fig = f"{self.workdir:s}/FF_performance_{gen_prefix:s}.png"
        errs = self.parameters.plot_ff_results(performance_fig, ref_dics, [
                                               params_opt], labels=gen_prefix)

        param_fig = f"{self.workdir:s}/parameters_{gen_prefix:s}.png"
        self.parameters.plot_ff_parameters(param_fig, [params_opt])

        # Save parameters
        self.parameters.export_parameters(
            self.ff_parameters, params_opt, errs[0])
        self._prepare_chm_info(params_opt)
        # self.parameters.generate_report(ref_dics, params_opt)

        return N_added

    def _prepare_chm_info(self, params0, params1=None, folder="calc", suffix="pyxtal0"):
        """
        Prepar_chm_info with from the given params.

        Args:
            params0 (array or list): FF parameters array
            params1 (array or list): FF parameters array
            folder (str): folder path
            suffix (str): suffix of the temporary file

        Returns:
            atom_info
        """
        pwd = os.getcwd()
        os.chdir(self.workdir)
        if not os.path.exists(folder):
            os.mkdir(folder)
        suffix = folder + '/' + suffix

        # To remove the old pyxtal1 files
        if os.path.exists(suffix + ".rtf"):
            os.remove(suffix + ".rtf")
        if os.path.exists(suffix + ".prm"):
            os.remove(suffix + ".prm")

        ase_with_ff = self.parameters.get_ase_charmm(params0)
        ase_with_ff.write_charmmfiles(base=suffix)
        if params1 is not None:
            ase_with_ff = self.parameters.get_ase_charmm(params1)
            ase_with_ff.write_charmmfiles(base=suffix)
        os.chdir(pwd)

        # Info
        return ase_with_ff.get_atom_info()

    def get_label(self, i):
        if i < 10:
            folder = f"cpu00{i}"
        elif i < 100:
            folder = f"cpu0{i}"
        else:
            folder = f"cpu0{i}"
        return folder

    def print_matches(self, header=None):
        """
        Formatted output for the matched structures with xtal rep and eng rank
        """
        if self.rank == 0:
            all_engs = np.sort(np.array(self.engs))
            ranks = []
            xtals = []
            if self.ref_pxrd is not None:
                matches = sorted(
                    self.matches, key=lambda x: -x[4])  # similarity
            else:
                print(self.matches)
                matches = sorted(self.matches, key=lambda x: x[3]) # eng

            for match_data in matches:
                d1, match = None, None
                if self.ref_pxrd is not None:
                    (_, id, xtal, e, match, tag) = match_data
                    add = self.new_struc(xtal, xtals)
                    if add:
                        xtals.append(xtal)
                else:
                    (_, id, xtal, e, d1, d2, tag) = match_data
                    add = True

                if add:
                    rep0 = xtal.get_1D_representation()
                    if header is not None:
                        strs = header
                    else:
                        strs = ""
                    strs += rep0.to_string(eng=xtal.energy / sum(xtal.numMols))
                    if d1 is not None:
                        strs += f"{d1:6.3f}{d2:6.3f} Match "
                    if match is not None:
                        strs += f" {match:4.2f} "

                    if e is not None:
                        rank = len(all_engs[all_engs < (e - 1e-3)]) + 1
                        strs += f" {rank:d}/{self.N_struc:d} {tag:s}"
                        ranks.append(rank)

                    print(strs)
            if len(ranks) == 0:
                ranks = [0]
            return min(ranks)

    def _print_match(self, xtal, ref_pmg):
        """
        Print the matched structure

        Args:
            rep: 1d rep
            eng: energy
            ref_pmg: reference pmg structure
        """

        rep0 = xtal.get_1D_representation()
        pmg_s1 = xtal.to_pymatgen()
        pmg_s1.remove_species("H")
        strs = rep0.to_string(eng=xtal.energy / sum(xtal.numMols))
        rmsd = self.matcher.get_rms_dist(ref_pmg, pmg_s1)
        if rmsd is not None:
            strs += f"{rmsd[0]:6.3f}{rmsd[1]:6.3f} Match Ref"
            print(strs)
            return rmsd[0], rmsd[1]
        else:
            return None, None

    def _apply_gaussian(self, reps, engs, h1=0.1, h2=0.1, w1=0.2, w2=3):
        """
        Apply Gaussian to discourage the sampling of already visited configs.
        Consider both lattice abc and torsion
        """
        from copy import deepcopy

        # check torsion
        N_id = 8
        engs_gau = deepcopy(engs)
        for i, rep in enumerate(reps):
            gau = 0
            if rep is not None and engs[i] < 9999:
                sg1, abc1 = rep[0][0], np.array(rep[0][1:])
                tor1 = np.zeros(self.N_torsion)
                count = 0
                for j in range(1, len(rep)):
                    if len(rep[j]) > N_id:  # for Cl-
                        tor1[count: count +
                             len(rep[j]) - N_id - 1] = rep[j][N_id:-1]
                        count += len(rep[j]) - N_id

                for ref in self.best_reps:
                    sg2, abc2 = ref[0][0], np.array(ref[0][1:])
                    # Cell
                    g1 = 0
                    if sg1 == sg2:
                        diff1 = np.sum((abc1 - abc2) ** 2) / w1**2
                        g1 = h1 * np.exp(-0.5 * diff1)  # cell
                    # Torsion
                    g2 = 0
                    if len(tor1) > 0:
                        tor2 = np.zeros(self.N_torsion)
                        count = 0
                        for j in range(1, len(rep)):
                            if len(rep[j]) > N_id:  # for Cl-
                                tor2[count: count +
                                     len(ref[j]) - N_id - 1] = ref[j][N_id:-1]
                                count += len(ref[j]) - N_id

                        diff2 = np.sum((tor1 - tor2) ** 2) / w2**2
                        g2 = h2 * np.exp(-0.5 * diff2)  # torsion
                    gau += g1 + g2
                # if gau > 1e-2: print(sg1, diag1, abc1, tor1)
            # print("Gaussian", i, "eng", engs[i], "gau", gau)
            # import sys; sys.exit()
            engs_gau[i] += gau
        return np.array(engs_gau)

    def check_ref(self, reps=None, reference=None, filename="pyxtal.cif"):
        """
        Check if ground state structure is found.

        Args:
            reps: list of representations
            refernce: [pmg, eng]
            filename: filename
        """
        if os.path.exists(filename):
            os.remove(filename)

        if reference is not None:
            [pmg0, eng] = reference
            pmg0.remove_species("H")
            print("check if ground state structure is found")

            if reps is None:
                reps = np.array(self.reps)

            if eng is None:
                eng = np.min(reps[:, -1]) + 0.25

            reps = reps[reps[:, -1] < (eng + 0.1)]
            ids = np.argsort(reps[:, -1])
            reps = reps[ids]
            new_reps = []
            for rep in reps:
                eng1 = rep[-1]
                rep0 = representation(rep[:-1], self.smiles)
                xtal = rep0.to_pyxtal()
                pmg_s1 = xtal.to_pymatgen()
                pmg_s1.remove_species("H")
                new = True
                for ref in new_reps:
                    eng2 = ref[-1]
                    pmg_s2 = representation(
                        rep[:-1], self.smiles).to_pyxtal().to_pymatgen()
                    pmg_s2.remove_species("H")
                    if abs(eng1 - eng2) < 1e-2 and self.matcher().fit(pmg_s1, pmg_s2):
                        new = False
                        break
                if new:
                    new_reps.append(rep)
                    header = f"{len(new_reps):d}: {eng1:12.4f}"
                    xtal.to_file(filename, header=header, permission="a+")
                    strs = rep0.to_string(eng=eng1)
                    rmsd = self.matcher.get_rms_dist(pmg0, pmg_s1)
                    if rmsd is not None:
                        strs += f"{rmsd[0]:6.3f}{rmsd[1]:6.3f} True"
                        print(strs)
                        return True
                    else:
                        print(strs)
        return False

    def _get_local_optimization_args(self):
        args = [
            randomizer,
            optimizer,
            self.smiles,
            self.block,
            self.num_block,
            self.atom_info,
            self.workdir + "/" + "calc",
            self.sg,
            self.composition,
            self.lattice,
            self.torsions,
            self.molecules,
            self.sites,
            self.ref_pmg,
            self.matcher,
            self.ref_pxrd,
            self.use_hall,
            self.skip_ani,
            self.check_stable,
            self.pre_opt,
        ]
        return args

    def local_optimization(self, xtals, qrs=False, pool=None):
        """
        Perform MPI optimization for each structure in each generation.

        Args:
            xtals : list of (xtal, tag) tuples
            qrs (bool): Force mutation or not (related to QRS)
        """
        if self.use_mpi:
            return self.local_optimization_mpi(xtals, qrs=qrs, pool=pool)
        elif self.ncpu == 1:
            return self.local_optimization_serial(xtals, qrs)
        else:
            print(f"Local optimization by multi-threads {self.ncpu}")
            return self.local_optimization_mproc(xtals, self.ncpu, qrs=qrs, pool=pool)

    def local_optimization_serial(self, xtals, qrs=False):
        """
        Perform optimization for each structure in each generation.

        Args:
            xtals : list of (xtal, tag) tuples
            qrs (bool): Force mutation or not (related to QRS)
        """
        args = self._get_local_optimization_args()
        gen = self.generation
        gen_results = [(None, None, None)] * len(xtals)
        for pop in range(len(xtals)):
            xtal = xtals[pop][0]
            job_tag = self.tag + "-g" + str(gen) + "-p" + str(pop)
            mutated = False if qrs else xtal is not None
            my_args = [xtal, pop, mutated, job_tag, *args]
            xtal, match = optimizer_single(*tuple(my_args))
            gen_results[pop] = (pop, xtal, match)
        return gen_results

    def local_optimization_mpi(self, xtals, qrs, pool):
        """
        Perform MPI optimization for each structure in each generation.

        Args:
            xtals : list of (xtal, tag) tuples
            qrs (bool): Force mutation or not (related to QRS)
        """
        #t0 = time()
        gen = self.generation
        self.print("Local optimization enabled by MPI", self.size, self.ncpu)

        # Distribute args_lists across available ranks (processes)
        local_xtals = xtals[self.rank::self.size]

        local_ids = list(range(self.N_pop))[self.rank::self.size]

        # Call local_optimization_mproc
        self.logging.info(f"Rank {self.rank} gets {len(local_xtals)} strucs")
        results = self.local_optimization_mproc(local_xtals,
                                                self.ncpu,
                                                local_ids,
                                                qrs,
                                                pool)
        # Synchronize before gathering
        self.logging.info(f"Rank {self.rank} finish local_optimization_mproc")
        self.comm.Barrier()

        # Gather all results at the root process
        self.logging.info(f"Rank {self.rank} in MPI_Gather at gen {gen}")
        all_results = self.comm.gather(results, root=0)
        self.logging.info(f"Rank {self.rank} done MPI_Gather at gen {gen}")

        # If root process, process the results
        gen_results = None
        if self.rank == 0:
            gen_results = [(None, None, None)] * len(xtals)
            for result_set in all_results:
                for res in result_set:
                    (id, xtal, match) = res
                    gen_results[id] = (id, xtal, match)

        # Broadcast
        self.logging.info(f"Rank {self.rank} MPI_bcast at gen {gen}")
        gen_results = self.comm.bcast(gen_results, root=0)

        return gen_results

    def local_optimization_mproc(self, xtals, ncpu, ids=None, qrs=False, pool=None):
        """
        Perform optimization for each structure in multiprocess mode.

        Args:
            xtals : list of (xtal, tag) tuples
            ncpu (int): number of parallel python processes
            ids (list): list of ids of the associated xtals
            qrs (bool): Force mutation or not (related to QRS)
        """
        gen = self.generation
        t0 = time()
        args = self._get_local_optimization_args()
        if ids is None:
            ids = range(len(xtals))

        N_cycle = int(np.ceil(len(xtals) / ncpu))
        # Generator to create arg_lists for multiprocessing tasks
        def generate_args_lists():
            for i in range(ncpu):
                id1 = i * N_cycle
                id2 = min([id1 + N_cycle, len(xtals)])
                _ids = ids[id1: id2]
                job_tags = [self.tag + "-g" + str(gen)
                            + "-p" + str(id) for id in _ids]
                _xtals = [xtals[id][0] for id in range(id1, id2)]
                mutates = [False if qrs else xtal is not None for xtal in _xtals]
                my_args = [_xtals, _ids, mutates, job_tags, *args, self.rank, self.timeout]
                yield tuple(my_args)  # Yield args instead of appending to a list

        gen_results = []
        for result in pool.imap_unordered(process_task, generate_args_lists()):
            if result is not None:
                for _res in result:
                    gen_results.append(_res)

        return gen_results

    def gen_summary(self, t0, gen_results, xtals):
        """
        Write the generic summary for each generation.

        Args:
            t0 (float): time stamp
            gen_results: list of results (id, xtal, match)
            xtals: list of (xtal, tag) tuples
        """
        matches = [False] if self.ref_pxrd is None else [0.0]
        matches *= self.N_pop

        eng0s = [self.E_max] * self.N_pop
        reps = [None] * self.N_pop
        new_xtals = [(None, None)] * len(xtals)
        gen = self.generation

        for res in gen_results:
            (id, xtal, match) = res

            if xtal is not None:
                new_xtals[id] = (xtal, xtals[id][1])
                eng0s[id] = xtal.energy / sum(xtal.numMols)
                reps[id] = xtal.get_1D_representation()
                matches[id] = match

                # Don't write bad structure
                if self.cif is not None and xtal.energy < 9999:
                    if self.verbose:
                        print("Add qualified structure", id, xtal.energy)
                    with open(self.cif, "a+") as f:
                        label = self.tag + "-g" + str(gen) + "-p" + str(id)
                        f.writelines(xtal.to_file(header=label))
                self.engs.append(xtal.energy / sum(xtal.numMols))
                self.stats[gen][id][0] = xtal.energy / sum(xtal.numMols)
                self.stats[gen][id][1] = match

        self.min_energy = np.min(np.array(self.engs))
        self.N_struc = len(self.engs)
        strs = f"Generation {gen:d} finishes: {len(self.engs):d} strucs"
        print(strs)
        self.logging.info(strs)

        t1 = time()

        # Apply Gaussian
        reps_x = [rep.x if rep is not None else None for rep in reps]
        if self.ref_pxrd is None:
            engs = self._apply_gaussian(reps_x, eng0s)
        else:
            engs = self._apply_gaussian(reps_x, -1 * np.array(matches))

        # Store the best structures
        count = 0
        ref_xtals = []
        ids = np.argsort(engs)
        for id in ids:
            (xtal, tag) = new_xtals[id]
            rep, eng = reps[id], eng0s[id]
            if self.new_struc(xtal, ref_xtals):
                ref_xtals.append(xtal)
                self.best_reps.append(rep.x)
                # d_rep = representation(rep, self.smiles)
                tag = f"{tag:8s}"
                try:
                    strs = rep.to_string(None, eng, tag)
                    out = f"{gen:3d} {strs:s} Top"
                    if self.ref_pxrd is not None:
                        out += f" {matches[id]:6.3f}"
                    print(out)
                except:
                    print('Error', xtal)
                count += 1
            if count == 3:
                break

        t2 = time()
        gen_out = f"Gen{gen:3d} time usage: "
        gen_out += f"{t1 - t0:5.1f}[Calc] {t2 - t1:5.1f}[Proc]"
        print(gen_out)

        return new_xtals, matches, engs

    def plot_results(self, save=True, figsize=(8.0, 5.0), figname=None, ylim=None):
        """
        Plot the results

        Args:
            save (bool): whether or not save the data
            figsize: e.g. (8.5, 5.0)
            figname (str):
            ylim: e.g. (0, 1.0)
        """
        if self.rank == 0:
            import matplotlib.pyplot as plt
            import seaborn as sns
            sns.set_theme()
            sns.set_context("talk", font_scale=0.9)

            if figname is None:
                figname = self.suffix + "-results.pdf"

            data1 = []
            data2 = []
            # Extract (pop_id, eng, gen_id) when pxrd is False
            # Extract (pop_id, eng, sim) when pxrd is True
            for i in range(self.N_gen):
                for j in range(self.N_pop):
                    if self.ref_pxrd is not None:
                        data1.append(
                            [i, j, self.stats[i, j, 0], self.stats[i, j, 1]])
                    else:
                        data1.append([i, j, self.stats[i, j, 0]])

            for match in self.matches:
                if self.ref_pxrd is not None:
                    data2.append([match[0], match[1], match[3], match[4]])
                else:
                    data2.append([match[0], match[1], match[3]])

            fig = plt.figure(figsize=figsize)
            plt.ylabel("Lattice Energy (kcal)")  # , weight='bold')
            data1 = np.array(data1)
            if self.ref_pxrd is not None:
                # (similarity, eng, gen_id)
                x1, y1, z1 = data1[:, 3], data1[:, 2], data1[:, 0]
                plt.xlabel("XRD Similarity")  # , weight='bold')
            else:
                x1, y1, z1 = data1[:, 1], data1[:, 2], data1[:, 0]
                plt.xlabel("Population ID")  # , weight='bold')
            # Plot of all samples (PopID, Engs/Similarity)
            scatter = plt.scatter(x1, y1, s=10, c=z1,
                                  cmap='winter', alpha=0.5, label='Samples')
            cbar = plt.colorbar(scatter)
            cbar.set_label('Generation ID')  # , weight='bold')
            if ylim is None:
                y1 = np.array(y1)
                ymin = y1.min() - 0.25
                ymax = min([ymin + 10.0, y1.max()])
                ylim = (ymin, ymax)

            if len(data2) > 0:
                data2 = np.array(data2)
                if len(data2.shape) == 1:
                    data2 = data2.reshape(-1, 1)
                if self.ref_pxrd is not None:
                    x2, y2, z2 = data2[:, 3], data2[:, 2], data2[:, 0]
                else:
                    x2, y2, z2 = data2[:, 1], data2[:, 2], data2[:, 0]
                plt.scatter(x2, y2, s=10, c='red', marker='x', label='Matches')

            plt.legend(loc=1)  # , prop={'weight': 'bold'})
            plt.ylim(ylim)
            plt.title(f"{self.name:s}-{self.ff_style:s}")  # , weight='bold')
            plt.tight_layout()
            plt.savefig(figname)

            if save:
                if self.ref_pxrd is not None:
                    header = "#Generation, Population, Energy, Similarity"
                else:
                    header = "#Generation, Population, Energy"
                data_txt = self.suffix + "-data.txt"
                np.savetxt(data_txt, data1, header=header)

                if len(data2) > 0:
                    match_txt = self.suffix + "-match.txt"
                    np.savetxt(match_txt, data2, header=header)


if __name__ == "__main__":
    print("test")



"""
some utilities
"""

import os
import re

import numpy as np
from ase import Atoms
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer as sga
from spglib import get_symmetry_dataset

from pyxtal.symmetry import Hall


def find_dir(dirs):
    """
    a short function to find the correct dir from a list
    """
    for d in dirs:
        if os.path.isdir(d):
            return d
    raise RuntimeError("Cannot find the dirtory for dftb parameters")


def listToString(s):
    # initialize an empty string
    # str1 = " "
    str1 = ""
    # return string
    return str1.join(s)


def pymatgen2ase(struc):
    """
    A short cut to convert between pymatgen to ase
    """
    atoms = Atoms(symbols=struc.atomic_numbers,
                  cell=struc.lattice.matrix, pbc=True)
    atoms.set_scaled_positions(struc.frac_coords)
    return atoms


def ase2pymatgen(struc):
    """
    A short cut to convert between pymatgen to ase
    """
    lattice = struc.cell
    coordinates = struc.get_scaled_positions()
    species = struc.get_chemical_symbols()
    return Structure(lattice, species, coordinates)


def symmetrize_cell(struc, mode="C"):
    """
    symmetrize structure from pymatgen, and return the struc in conventional or
    primitive setting.

    Args:
        struc: ase type
        mode: output conventional or primitive cell
    """
    P_struc = ase2pymatgen(struc)
    finder = sga(P_struc, symprec=0.06)
    P_struc = finder.get_conventional_standard_structure(
    ) if mode == "C" else finder.get_primitive_standard_structure()

    return pymatgen2ase(P_struc)


def good_lattice(struc, maxvec=50.0, minvec=1.2, maxang=150, minang=30):
    """
    check if the lattice has a good shape.

    Args:
        struc: pyxtal structure
    """

    para = struc.lattice.get_para(degree=True)
    return bool(max(para[:3]) < maxvec and min(para[:3]) > minvec and max(para[3:]) < maxang and min(para[3:]) > minang)


def symmetrize(pmg, tol=1e-3, a_tol=5.0, style="pyxtal", hn=None):
    """
    symmetrize the structure from spglib.

    Args:
        pmg: pymatgen structure
        tol: tolerance
        a_tol: angle tolerance
        style: 'pyxtal' or spglib, differing in the choice of origin
        hn: hall_number

    Returns:
        pymatgen structure with symmetrized lattice
    """
    numbers = [site.species.elements[0].Z for site in pmg.sites]
    atoms = (pmg.lattice.matrix, pmg.frac_coords, numbers)
    dataset = get_symmetry_dataset(atoms, tol, angle_tolerance=a_tol)
    if hn is None:
        hn = Hall(dataset.number, style=style).hall_default
    if hn != dataset.hall_number:
        dataset = get_symmetry_dataset(
            atoms, tol, angle_tolerance=a_tol, hall_number=hn)
    cell = dataset.std_lattice
    pos = dataset.std_positions
    numbers = dataset.std_types

    return Structure(cell, numbers, pos), hn


def get_symmetrized_pmg(pmg, tol=1e-3, a_tol=5.0, style="pyxtal", hn=None):
    """
    Get the symmetrized Pymatgen structure. A slight modification to ensure that
    the structure adopts the standard setting according to the Interational
    Crystallography Table.

    Args:
        pmg: input pymatgen structure
        tol: symmetry tolerance
        a_tol: angle tolerance
        style: 'pyxtal' or spglib, differing in the choice of origin
        hn: hall_number

    Returns:
        pymatgen structure with symmetrized lattice
    """

    pmg, hn = symmetrize(pmg, tol, a_tol=a_tol, style=style, hn=hn)
    s = sga(pmg, symprec=tol, angle_tolerance=a_tol)
    # make sure that the coordinates are in standard setting
    # if hn is None:
    #    hn = Hall(s._space_group_data['number'], style=style).hall_default
    if hn != s._space_group_data.hall_number:
        s._space_group_data = get_symmetry_dataset(
            s._cell, tol, angle_tolerance=a_tol, hall_number=hn)
    return s.get_symmetrized_structure(), s.get_space_group_number()


def extract_ase_db(db_file, id):
    """
    a short cut to extract the structural information
    from the ase db file by row id
    """

    import os

    from ase.db import connect

    from pyxtal import pyxtal

    if not os.path.exists("output"):
        os.makedirs("output")
    print("Dumping the structures to the folder output")
    with connect(db_file) as db:
        for id in ids:
            s = db.get_atoms(id=id)
            filename = "output/" + str(id) + ".vasp"
            s.write(filename, format="vasp", direct=True, vasp5=True)
            my = pyxtal()
            my.from_seed(s)
            print(my)


def parse_cif(filename, header=False, spg=False, eng=False, csd=False, sim=False):
    """
    read structures from a cif (our own format with #END)
    Args:
        filename: string
        header: bool, whether or not return header
        spg: bool, whether or not return the spg
    """
    strings = []
    headers = []
    spgs = []
    engs = []
    csds = []
    sims = []
    with open(filename) as f:
        lines = f.readlines()
        start = None
        end = None
        for i in range(len(lines)):
            if lines[i].find("data_") == 0:
                if sim:
                    sims.append(float(lines[i].split(":")[-1]))
                end = i
                if start is not None:
                    tmp = []
                    for l in lines[start: end - 1]:
                        if len(re.findall(r"[0-9][B-C]", l)) > 0 or len(re.findall(r"[A-Z][0-9]\' [0-9]", l)) > 0:
                            # print(l) #; import sys; sys.exit()
                            continue
                        tmp.append(l)
                    cif = listToString(tmp)
                    strings.append(cif)
                start = i
                headers.append(lines[i])
            elif lines[i].find("_symmetry_Int_Tables_number") == 0:
                spgs.append(int(lines[i].split()[-1]))
            elif lines[i].find("#Energy") == 0:
                engs.append(float(lines[i].split()[1]))
            elif lines[i].find("_database_code") == 0:
                tmp = lines[i].split()[-1]
                csds.append(tmp[:-1])

        # Last one
        tmp = []
        for l in lines[start:]:
            if len(re.findall(r"[0-9][B-D]", l)) > 0 or len(re.findall(r"[A-Z][0-9]\' [0-9]", l)) > 0:
                # print(l);
                continue
            tmp.append(l)
        cif = listToString(tmp)
        strings.append(cif)

    if header:
        return strings, headers
    elif spg:
        return strings, spgs
    elif eng:
        return strings, engs
    elif csd:
        return strings, csds
    elif sim:
        return strings, sims
    else:
        return strings


def process_csd_cif(cif, remove_H=False):
    """
    process cif from CSD, sometimes it contains multiple
    e.g., C2
    """
    lines = cif.split("\n")
    tmp = []
    for l in lines:
        if (
            len(re.findall(r"[0-9][A-Z]", l)) > 0
            or len(re.findall(r"[A-Z]\?", l)) > 0
            or len(re.findall(r"[0-9]\?", l)) > 0
            or len(re.findall(r"[A-Z][0-9]\' [0-9]", l)) > 0
        ):
            # len(re.findall(r"[0-9]_[0-9]", l))>0 or \
            # print(l) #; import sys; sys.exit()
            continue

        if remove_H and len(re.findall(r" H ", l)) > 0:
            continue

        tmp.append(l + "\n")
    return listToString(tmp)


def get_similar_cids_from_pubchem(base, MaxRecords):
    """
    Args:
        base: PubChem CID of Starting chemical
        MaxRecords: Number of Similar Compounds

    Returns:
        List of the CIDs of PubChem compounds similar to the base compound.
    """

    import pubchempy as pcp

    if isinstance(base, int):
        base = str(base)
    cids = pcp.get_compounds(
        base, searchtype="similarity", MaxRecords=MaxRecords)
    results = []
    for x in cids:
        csd_codes = search_ccdc_structures(x.cid)
        if len(csd_codes) > 0:
            d = {
                "cid": x.cid,
                "smiles": x.canonical_smiles,
                "name": x.iupac_name,
                "csd_codes": csd_codes,
            }
            results.append(d)
            print(d)
    return results


def search_csd_code_by_pubchem(cid):
    """
    Args:
        cid: PubChem cid

    Returns:
        CIDs that have CCDC crystal structure data
    """

    import json
    import urllib

    from monty.json import MontyDecoder

    url0 = "https://pubchem.ncbi.nlm.nih.gov/rest/pug_view/data/compound/"
    cid = str(cid)
    url = url0 + cid + "/JSON"

    try:
        response = urllib.request.urlopen(url)
    except urllib.error.HTTPError:
        print("Problem in http connection", url)
        return None
    except urllib.error.URLError:
        print("Problem in parsing", url)
        return None

    try:
        contents = response.read()
        if contents.decode().find("CCDC") > 0:
            data = json.loads(contents, cls=MontyDecoder)
            if ("Section" in data["Record"]["Section"][0]) & (len(data["Record"]["Section"][0]["Section"]) == 3):
                infos = data["Record"]["Section"][0]["Section"][2]["Section"][0]["Information"]
                csd_codes = [info["Value"]["StringWithMarkup"]
                             [0]["String"] for info in infos]
    except:
        csd_codes = []
        print("Failed to parse json", url, "\n")

    return csd_codes


def search_csd_entries_by_code(code):
    """
    Args:
        code: CSD code, e.g., ACSALA

    Returns:
        list of csd ids
    """

    from ccdc.crystal import PackingSimilarity as PS
    from ccdc.search import TextNumericSearch

    def new_cryst(cryst, crysts, n_max):
        spg1 = cryst.spacegroup_number_and_setting[0]
        for ref in crysts:
            spg2 = ref.spacegroup_number_and_setting[0]
            if spg1 == spg2:
                h = PS().compare(cryst, ref)
                # print(cryst.identifier, ref.identifier, h.nmatched_molecules)
                if h is not None and h.nmatched_molecules == n_max:
                    return False
        return True

    n_max = PS().settings.packing_shell_size
    query = TextNumericSearch()
    query.add_identifier(code)
    hits = query.search()
    unique_crysts = []
    for hit in hits:
        if (hit.entry.has_3d_structure and hit.entry.pressure is None) & (new_cryst(hit.crystal, unique_crysts, n_max)):
            unique_crysts.append(hit.crystal)
            # print(hit.entry.identifier, hit.entry.deposition_date)

    return [c.identifier for c in unique_crysts]


def get_struc_from__parser(p):
    """
    A utility to get the pymatgen structure from the CifParser
    Sometimes the cif structure may have repeated atom entries

    Args:
        p: pymatgen CifParser object

    Return:
        a single pymatgen structure
    """
    from collections import OrderedDict

    import numpy as np
    from pymatgen.core.periodic_table import get_el_sp
    from pymatgen.io.cif import str2float
    from pymatgen.util.coord import find_in_coord_list_pbc

    def get_matching_coord(coord, ops, atol=1e-4):
        keys = list(coord_to_species.keys())
        coords = np.array(keys)
        for op in ops:
            c = op.operate(coord)
            inds = find_in_coord_list_pbc(coords, c, atol=atol)
            if len(inds):
                return keys[inds[0]]
        return False

    for i, d in enumerate(p._cif.data.values()):
        ops = p.get_symops(d)
        coord_to_species = OrderedDict()

        d0 = {
            "_atom_site_label": [],
            "_atom_site_fract_x": [],
            "_atom_site_fract_y": [],
            "_atom_site_fract_z": [],
        }

        for j in range(len(d["_atom_site_label"])):
            try:
                symbol = p._parse_symbol(d["_atom_site_type_symbol"][j])
            except KeyError:
                symbol = p._parse_symbol(d["_atom_site_label"][j])

            el = get_el_sp(symbol)
            x = str2float(d["_atom_site_fract_x"][j])
            y = str2float(d["_atom_site_fract_y"][j])
            z = str2float(d["_atom_site_fract_z"][j])

            coord = (x, y, z)
            match = get_matching_coord(coord, ops)
            if not match:
                d0["_atom_site_label"].append(el)
                d0["_atom_site_fract_x"].append(str(x))
                d0["_atom_site_fract_y"].append(str(y))
                d0["_atom_site_fract_z"].append(str(z))
                coord_to_species[coord] = el

        d.data["_atom_site_label"] = d0["_atom_site_label"]
        d.data["_atom_site_fract_x"] = d0["_atom_site_fract_x"]
        d.data["_atom_site_fract_y"] = d0["_atom_site_fract_y"]
        d.data["_atom_site_fract_z"] = d0["_atom_site_fract_z"]

        return p._get_structure(d, primitive=False, symmetrized=False)
    return None


def Kgrid(struc, Kresol=0.10, dimension=3):
    """
    Assign kpoints based on the lattice
    """
    # a, b, c, alpha, beta, gamma = struc.get_cell_lengths_and_angles()
    a, b, c, alpha, beta, gamma = struc.cell.cellpar()
    vol = struc.get_volume()
    dist = np.zeros(3)
    dist[2] = np.abs(vol / (a * b * np.sin(np.radians(gamma))))
    dist[1] = np.abs(vol / (a * c * np.sin(np.radians(beta))))
    dist[0] = np.abs(vol / (b * c * np.sin(np.radians(alpha))))
    Kpoints = np.ceil(1.0 / (dist * Kresol))
    if dimension == 2:
        Kpoints[-1] = 1
    # print(a, b, c, alpha, beta, gamma)
    # print(vol/(a*b*np.sin(gamma)), a*b, np.sin(gamma))
    # print(vol/(a*c*np.sin(beta)), a*c, np.sin(beta))
    # print(vol/(b*c*np.sin(alpha)), b*c, np.sin(alpha))
    # print(Kpoints)
    # import sys; sys.exit()
    return Kpoints.astype(int)


def sort_by_dimer(atoms, N_mols, id=10, tol=4.0):
    """
    sort the ase atoms' xyz according to dimer
    so far only tested on aspirin

    Args:
        atoms: atoms object from pyxtal
        N_mols: number of molecules
        id: the refrence atom id
        tol: tolerence distance to check if it is a dimer
    """

    N_atoms = int(len(atoms) / N_mols)
    pos = atoms.get_scaled_positions()
    refs = pos[id: len(pos): N_atoms, :]
    # print(refs)

    # compuate the indices and shift
    orders = []
    shifts = []
    while len(orders) < N_mols:
        lefts = [i for i in range(N_mols) if i not in orders]
        i = lefts[0]
        orders.append(i)
        shifts.append(np.zeros(3))
        ref_i = refs[i]
        good = False
        for j in lefts[1:]:
            ref_j = refs[j]
            dist = ref_j - ref_i
            shift = np.rint(dist)
            dist -= shift
            dist = np.linalg.norm(dist.dot(atoms.cell[:]))
            if dist < tol:
                orders.append(j)
                shifts.append(shift)
                good = True
                break
        if not good:
            raise RuntimeError("Cannot find match on molecule", i)

        print("get", i, j, dist, shift)

    pos0 = atoms.get_positions()
    pos1 = np.zeros([len(pos), 3])
    for i, id in enumerate(orders):
        s1, e1 = id * N_atoms, (id + 1) * N_atoms
        s2, e2 = i * N_atoms, (i + 1) * N_atoms
        pos1[s2:e2, :] += pos0[s1:e1, :] - shifts[i].dot(atoms.cell[:])

    atoms.set_positions(pos1)
    return atoms


def generate_wp_lib(
    spg_list,
    composition,
    num_wp=(None, None),
    num_fu=(None, None),
    num_dof=(None, None),
    N_max=1000,
):
    """
    Generate wps according to the composition constraint (e.g., SiO2)

    Args;
        - spg_list: list of space group choices
        - composition: chemical compositions [1, 2]
        - num_wp: (min_wp, max_wp)
        - num_fu: (min_fu, max_fu)
        - num_dof: (min_dof, max_dof)

    Returns:
        a list of wps [spg, ([wp1, ...], ... [wp1, ...]), dof]
    """

    from pyxtal.symmetry import Group

    composition = np.array(composition, dtype=int)
    (min_wp, max_wp) = num_wp
    (min_fu, max_fu) = num_fu
    (min_dof, max_dof) = num_dof

    if max_wp is None:
        max_wp = len(composition)
    if min_wp is None:
        min_wp = len(composition)
    if min_dof is None:
        min_dof = 1
    if max_dof is None:
        max_dof = 1000
    # print(max_wp, min_wp)
    wps = []
    for sg in spg_list:
        g = Group(sg)
        lat_dof = g.get_lattice_dof()
        # determine the upper and lower limit
        if min_fu is None:
            min_fu = max([int(len(g[-1]) / min(composition)), 1])
        if max_fu is None:
            max_fu = max([int(len(g[0]) / max(composition)), 1])
        count = 0
        for i in range(max_fu, min_fu - 1, -1):
            letters, _, wp_ids = g.list_wyckoff_combinations(
                composition * i, numWp=(min_wp, max_wp), Nmax=100000)
            for _j, wp in enumerate(wp_ids):
                wp_dofs = 0
                num = 0
                for wp0 in wp:
                    for id in wp0:
                        wp_dofs += g[id].get_dof()
                        num += g[id].multiplicity
                # print(sg, wp, letters[j])
                num_dof = lat_dof + wp_dofs
                if min_dof <= num_dof <= max_dof:
                    wps.append((num, sg, wp, lat_dof + wp_dofs))
                    count += 1
            if count >= N_max:
                break
    return wps


def reset_lammps_cell(atoms0):
    """
    set the cell into lammps format
    """
    from ase.calculators.lammpslib import convert_cell

    atoms = atoms0.copy()
    mat, coord_transform = convert_cell(atoms0.cell)
    if coord_transform is not None:
        pos = np.dot(atoms0.positions, coord_transform.T)
        atoms.set_cell(mat.T)
        atoms.set_positions(pos)
    return atoms


def new_struc(xtal, xtals):
    """
    check if this is a new structure

    Args:
        xtal: input structure
        xtals: list of reference structures

    Return:
        `None` or the id of matched structure
    """
    import pymatgen.analysis.structure_matcher as sm

    if xtal is None or not hasattr(xtal, "energy"):
        return False
    else:
        eng1 = xtal.energy
        pmg_s1 = xtal.to_pymatgen()
        pmg_s1.remove_species("H")
        vol1 = pmg_s1.lattice.volume

        for xtal2 in xtals:
            # print(rep); print(rep2)
            eng2 = xtal2.energy
            if abs(eng1 - eng2) < 1e-2:
                pmg_s2 = xtal2.to_pymatgen()
                vol2 = pmg_s2.lattice.volume
                if abs(vol1 - vol2) / vol1 < 5e-2:
                    pmg_s2.remove_species("H")
                    if sm.StructureMatcher().fit(pmg_s1, pmg_s2):
                        # print(pmg_s2); import sys; sys.exit()
                        return False
        return True


def new_struc_wo_energy(xtal, xtals, ltol=0.2, stol=0.3, angle_tol=5.0, verbose=False):
    """
    check if this is a new structure

    Args:
        xtal: input structure
        xtals: list of reference structures
        ltol (float): Fractional length tolerance. Default is 0.2.
        stol (float): Site tolerance. ( V / Nsites ) ** (1/3). Default is 0.3.
        angle_tol (float): Angle tolerance in degrees. Default is 5 degrees.

    Return:
        `None` or the id of matched structure
    """
    import pymatgen.analysis.structure_matcher as sm

    if xtal is None:
        return False
    else:
        pmg_s1 = xtal.to_pymatgen()
        pmg_s1.remove_species("H")
        vol1 = pmg_s1.lattice.volume

        for xtal2 in xtals:
            if xtal.group.number == xtal2.group.number:
                # print(rep); print(rep2)
                pmg_s2 = xtal2.to_pymatgen()
                vol2 = pmg_s2.lattice.volume
                if abs(vol1 - vol2) / vol1 < 1e-2:
                    pmg_s2.remove_species("H")
                    if sm.StructureMatcher(ltol, stol, angle_tol).fit(pmg_s1, pmg_s2):
                        if verbose:
                            print(xtal, xtal2)  # ; import sys; sys.exit()
                        return False
        return True


def split_list_by_ratio(nums, ratio):
    """
    Splits a list of integers into two groups such that the sum of each group
    satisfies a given ratio and returns all possible ways of combinations
    tracking the indices of the numbers.

    Args:
        nums (list of int): The list of integers to split.
        ratio (tuple of int): A tuple of the desired ratio (e.g., (1, 1) for 1:1 ratio).

    Returns:
        list of tuple: A list of tuples where each contains two lists of indices.
                       Each pair of lists represents one possible way to split the
                       numbers to satisfy the given ratio.
    """

    def find_splits(i, sum1, sum2, group1, group2):
        if i == len(nums):
            if sum1 * ratio[1] == sum2 * ratio[0]:
                solutions.append((group1.copy(), group2.copy()))
            return

        # Try adding the current number's index to the first group
        group1.append(i)
        find_splits(i + 1, sum1 + nums[i], sum2, group1, group2)
        group1.pop()

        # Try adding the current number's index to the second group
        group2.append(i)
        find_splits(i + 1, sum1, sum2 + nums[i], group1, group2)
        group2.pop()

    solutions = []
    # Optional: Sorting can sometimes speed up the process
    nums = sorted(nums, reverse=True)

    find_splits(0, 0, 0, [], [])
    return solutions


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("-f", dest="file", help="path of database file")
    parser.add_argument(
        "-i",
        dest="id",
        help="index of the row",
    )

    options = parser.parse_args()
    ids = options.id
    ids = [int(id) for id in ids.split(",")] if ids.find(
        ",") > 0 else [int(ids)]
    extract_ase_db(options.file, ids)
